/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2021-2021. All rights reserved.
 * Author: bao
 */
#ifndef OCK_HCOM_12342434341233_H
#define OCK_HCOM_12342434341233_H

#include <stdbool.h>
#include <stdint.h>

#ifdef __cplusplus
extern "C" {
#endif

#define NUM_16 16
#define NUM_64 64
#define NUM_96 96
#define NUM_128 128
#define NUM_256 256
#define NUM_771 771
#define NUM_772 772
#define NUM_1024 1024

#define NET_SGE_MAX_IOV 4
#define NET_C_FLAGS_BIT(i) (1UL << (i))
#define NET_C_EP_SELF_POLLING NET_C_FLAGS_BIT(0)
#define NET_C_EP_EVENT_POLLING NET_C_FLAGS_BIT(1)

/*
 * @brief Driver, which include oob & rdma communication & callback etc
 */
typedef uintptr_t Net_Driver;

/*
 * @brief Endpoint, represent one RDMA connection to dual-direction communication
 *
 * two side operation, Net_EPPostSend
 * read operation from remote, Net_EPPostRead
 * write operation from remote, Net_EPPostWrite
 */
typedef uintptr_t Net_EndPoint;

/*
 * @brief MemoryRegion, which region memory in RDMA Nic for write/read operation
 */
typedef uintptr_t Net_MemoryRegion;

/*
 * @brief Two side RDMA operation (i.e. RDMA send/receive)
 *
 * @param data, pointer of data need to send to peer (the data will be copied to register RDMA memory region
 * the data must be less than mrSendReceiveSegSize of driver
 * @param size, size of data
 */
typedef struct {
    uintptr_t data;         // pointer of data to send to peer
    uint32_t size;          // size of data
    uint16_t upCtxSize;     // user context size
    char upCtxData[NUM_16]; // user context
} Net_SendRequest;

typedef struct {
    uint32_t seqNo;    // seq no
    int16_t timeout;  // timeout
    int16_t errorCode; // error code
    uint8_t flags;     // flags
} Net_OpInfo;

/*
 * @brief Device information for user
 */
typedef struct {
    int maxSge; // max iov count in NetTransSglRequest
} Net_DeviceInfo;

/*
 * @brief Read/write request for one side rdma operation
 */
typedef struct {
    uintptr_t lMRA;         // local memory region address
    uintptr_t rMRA;         // remote memory region address
    uint32_t lKey;          // local memory region key
    uint32_t rKey;          // remote memory region key
    uint32_t size;          // data size
    uint16_t upCtxSize;     // user context size
    char upCtxData[NUM_16]; // user context
} Net_ReadWriteRequest;

typedef struct {
    uintptr_t lAddress; // local memory region address
    uintptr_t rAddress; // remote memory region address
    uint32_t lKey;      // local memory region key
    uint32_t rKey;      // remote memory region key
    uint32_t size;      // data size
} __attribute__((packed)) Net_ReadWriteSge;

typedef struct {
    Net_ReadWriteSge *iov;  // sgl array
    uint16_t iovCount;      // max count:NUM_4
    uint16_t upCtxSize;     // user context size
    char upCtxData[NUM_16]; // user context
} __attribute__((packed)) Net_ReadWriteSglRequest;

/*
 * @brief Read/write mr info for one side rdma operation
 */
typedef struct {
    uintptr_t lAddress; // local memory region address
    uint32_t lKey;      // local memory region key
    uint32_t size;      // data size
} Net_MemoryRegionInfo;

/*
 * @brief Request type, part of Net_RequestContext
 */
typedef enum {
    C_SENT = 0,
    C_SENT_RAW = 1,
    C_SENT_RAW_SGL = 2,
    C_RECEIVED = 3,
    C_RECEIVED_RAW = 4,
    C_WRITTEN = 5,
    C_READ = 6,
    C_SGL_WRITTEN = 7,
    C_SGL_READ = 8,
} Net_RequestType;

/*
 * @brief Callback function context, for received, post done, read/write done
 */
typedef struct {
    Net_RequestType type;
    uint16_t opCode;   // for post send
    uint16_t flags;    // flags on the header
    int16_t timeout;  // timeout
    int16_t errorCode; // error code
    int result;        // return 0 successful
    void *msgData;     // for receive operation or C_OP_REQUEST_RECEIVED callback
    uint32_t msgSize;  // for receive operation or C_OP_REQUEST_RECEIVED callback
    uint32_t seqNo;    // for post send raw
    Net_EndPoint ep;
    Net_SendRequest originalSend;           // for C_OP_REQUEST_POSTED, copy struct information, not original
                                            // originalSend.data is self rdma address, not original input data address
    Net_ReadWriteRequest originalReq;       // for C_OP_READWRITE_DONE, copy struct information, not original
    Net_ReadWriteSglRequest originalSglReq; // for C_OP_READWRITE_DONE, copy struct information, not original
} Net_RequestContext;

typedef struct {
    uint16_t opCode;
    uint32_t seqNo;
    void *msgData;
    uint32_t msgSize;
} Net_ResponseContext;

typedef struct {
    uint32_t pid;
    uint32_t uid;
    uint32_t gid;
} Net_UdsIdInfo;

/*
 * @brief Worker polling type
 * 1 For RDMA:
 * C_BUSY_POLLING, means cpu 100% polling no matter there is request done, better performance but cost dedicated CPU
 * C_EVENT_POLLING, waiting on OS kernel for request done
 * 2 For TCP/UDS
 * only event pooling is supported
 */
typedef enum {
    C_BUSY_POLLING = 0,
    C_EVENT_POLLING = 1,
} Net_DriverWorkingMode;

typedef enum {
    C_DRIVER_RDMA = 0,
    C_DRIVER_TCP = 1,
    C_DRIVER_UDS = 2,
    C_DRIVER_SHM = 3,
} Net_DriverType;

/*
 * @brief DriverOobType working mode
 */
typedef enum {
    C_NET_OOB_TCP = 0,
    C_NET_OOB_UDS = 1,
} Net_DriverOobType;

/*
 * @brief Enum for secure type
 */
typedef enum {
    C_NET_SEC_DISABLED = 0,
    C_NET_SEC_ONE_WAY = 1,
    C_NET_SEC_TWO_WAY = 2,
} Net_DriverSecType;

typedef enum {
    C_TLS_1_2 = NUM_771,
    C_TLS_1_3 = NUM_772,
} Net_DriverTlsVersion;

typedef enum {
    C_OPENSSL = 0,
    C_HITLS = 1,
} Net_DriverTlsMode;
/*
 * @brief DriverCipherSuite mode
 */
typedef enum {
    C_AES_GCM_128 = 0,
    C_AES_GCM_256 = 1,
    C_AES_CCM_128 = 2,
    C_CHACHA20_POLY1305 = 3,
} Net_DriverCipherSuite;

/*
 * @brief Options for driver initialization
 */
typedef struct {
    Net_DriverWorkingMode mode;        // polling mode
    uint32_t mrSendReceiveSegCount;    // segment count of segment for send/receive
    uint32_t mrSendReceiveSegSize;     // single segment size of send/receive memory region
    char netDeviceIpMask[NUM_256];     // device ip mask, for multiple net device cases
    char netDeviceIpGroup[NUM_1024];   // ip group for devices
    uint16_t completionQueueDepth;     // rdma completion queue size
    uint16_t maxPostSendCountPerQP;    // max post send count
    uint16_t prePostReceiveSizePerQP;  // pre post receive size for one qp
    uint16_t pollingBatchSize;         // polling wc size on at one time
    uint32_t qpSendQueueSize;          // qp send queue size, by default is 256
    uint32_t qpReceiveQueueSize;       // qp receive queue size, by default is 256
    uint16_t dontStartWorkers;         // start worker or not, 1 means don't start, 0 means start
    char workerGroups[NUM_64];         // worker groups, for example 1,3,3
    char workerGroupsCpuSet[NUM_128];  // worker groups cpu set, for example 1-16
    // worker thread priority [-20,20], 20 is the lowest, -20 is the highest, 0 (default) means do not set priority
    int workerThreadPriority;
    uint16_t heartBeatIdleTime;        // heart beat idle time, in seconds
    uint16_t heartBeatProbeTimes;      // heart beat probe times, in seconds
    uint16_t heartBeatProbeInterval;   // heart beat probe interval, in seconds
    // timeout during io (s), it should be [-1, 1024], -1 means do not set, 0 means never timeout during io
    int16_t tcpUserTimeout;
    bool tcpEnableNoDelay;             // tcp TCP_NODELAY option, true in default
    bool tcpSendZCopy;                 // tcp whether copy request to inner memory, false in default
    /* The buffer sizes will be adjusted automatically when these two variables are 0, and the performance would be
     * better */
    uint16_t tcpSendBufSize;           // tcp connection send buffer size in kernel, in KB
    uint16_t tcpReceiveBufSize;        // tcp connection send receive buf size in kernel, in KB
    uint16_t enableTls;                // value only in 0 and 1, value 1 means enable ssl and encrypt, 0 on the contrary
    Net_DriverTlsMode tlsMode;         // tls mode, default is openssl
    Net_DriverSecType secType;         // security type
    Net_DriverTlsVersion tlsVersion;   // tls version, default TLS1.3 (772)
    Net_DriverCipherSuite cipherSuite; // if tls enabled can set cipher suite, client and server should same
    Net_DriverOobType oobType;         // oob type, tcp or UDS, UDS cannot accept remote connection
    uint8_t version;                   // program version used by connect validation
    uint32_t maxConnectionNum;         // max connection number
    char oobPortRange[NUM_16];         // port range when enable port auto selection
} Net_DriverOptions;

/*
 * @brief Options for multiple listeners
 */
typedef struct {
    char ip[NUM_16];            // ip to be listened
    uint16_t port;              // port to be listened
    uint16_t targetWorkerCount; // the count of workers can be dispatched to, for connections from this listener
} Net_DriverOobListenOptions;

/*
 * @brief Oob uds listening information
 */
typedef struct {
    char name[NUM_96];          /* UDS name for listen or file path */
    uint16_t perm;              /* if 0 means not use file, otherwise use file and this perm as file perm */
    uint16_t targetWorkerCount; /* the count of target workers, if >= 1, the
                                       accepted socket will be attached to sub set to workers, 0 means all */
} Net_OobUDSListenerOptions;

/*
 * @brief Callback function definition
 * 1) new endpoint connected from client, only need to register this at sever side
 * 2) endpoint is broken, called when RDMA qp detection error or broken
 */
typedef int (*Net_EPHandler)(Net_EndPoint ep, uint64_t usrCtx, const char *payLoad);

/*
 * @brief Callback function definition
 *
 * it is called when the following cases happen
 * 1) post send done
 * 2) read done
 * 3) write done
 *
 * Important notes:
 * 1) ctx is a thread local static variable, cannot transform to another thread directly
 * 2) msgData need to copy to another space properly
 * 3) ep can be transferred to another thread for further reply or other stuff
 * in this case, need to call Net_EPRefer() to increase reference count
 * and call Net_EPDestroy() after to decrease the reference count
 */
typedef int (*Net_RequestHandler)(Net_RequestContext *ctx, uint64_t usrCtx);

/*
 * @brief Idle callback function, when worker thread idle, this function will be called
 *
 * @param wkrGrpIdx        [in] worker group index in on net driver
 * @param idxInGrp         [in] worker index in the group
 * @param usrCtx           [in] user context
 */
typedef void (*Net_IdleHandler)(uint8_t wkrGrpIdx, uint16_t idxInGrp, uint64_t usrCtx);

/*
 * @brief Sec callback function, when oob connect build, this function will be called to generate auth info.
 * if this function not set secure type is C_NET_SEC_NO_VALID and oob will not send secure info
 *
 * @param ctx              [in] ctx from connect param ctx, and will send in auth process
 * @param flag             [out] flag to sent in auth process
 * @param type             [out] secure type, value should set in oob client, and should in [C_NET_SEC_ONE_WAY,
 * C_NET_SEC_TWO_WAY]
 * @param output           [out] secure info created
 * @param outLen           [out] secure info length
 * @param needAutoFree     [out] secure info need to auto free in hcom or not
 */
typedef int (*Net_SecInfoProvider)(uint64_t ctx, int64_t *flag, Net_DriverSecType *type, char **output,
    uint32_t *outLen, int *needAutoFree);

/*
 * @brief ValidateSecInfo callback function, when oob connect build, this function will be called to validate auth info
 * if this function not set oob will not validate secure info
 *
 * @param flag             [in] flag received in auth process
 * @param ctx              [in] ctx received in auth process
 * @param input            [in] secure info received
 * @param inputLen         [in] secure info length
 */
typedef int (*Net_SecInfoValidator)(uint64_t ctx, int64_t flag, const char *input, uint32_t inputLen);

/*
 * @brief Enum for tls callback, set peer cert verify type
 */
typedef enum {
    C_VERIFY_BY_NONE = 0,
    C_VERIFY_BY_DEFAULT = 1,
    C_VERIFY_BY_CUSTOM_FUNC = 2,
} Net_PeerCertVerifyType;

/*
 * @brief keyPass          [in] erase function
 * @param keyPass          [in] the memory address of keyPass
 */
typedef void (*Net_TlsKeyPassErase)(char *keyPass, int len);

/*
 * @brief The cert verify function
 *
 * @param x509             [in] the x509 object of CA
 * @param crlPath          [in] the crl file path
 *
 * @return -1 for failed, and 1 for success
 */
typedef int (*Net_TlsCertVerify)(void *x509, const char *crlPath);

/*
 * @brief Get the certificate file of public key
 *
 * @param name             [out] the name
 * @param certPath         [out] the path of certificate
 */
typedef int (*Net_TlsGetCertCb)(const char *name, char **certPath);

/*
 * @brief Get private key file's path and length, and get the keyPass
 * @param name             [out] the name
 * @param priKeyPath       [out] the path of private key
 * @param keyPass          [out] the keyPass
 * @param erase            [out] the erase function
 */
typedef int (*Net_TlsGetPrivateKeyCb)(const char *name, char **priKeyPath, char **keyPass, Net_TlsKeyPassErase *erase);

/*
 * @brief Get the CA and verify
 * @param name             [out] the name
 * @param caPath           [out] the path of CA file
 * @param crlPath          [out] the crl file path
 * @param verifyType       [out] the type of verify in[VERIFY_BY_NONE,VERIFY_BY_DEFAULT, VERIFY_BY_CUSTOM_FUNC]
 * @param verify           [out] the verify function, only effect in VERIFY_BY_CUSTOM_FUNC mode
 */
typedef int (*Net_TlsGetCACb)(const char *name, char **caPath, char **crlPath, Net_PeerCertVerifyType *verifyType,
    Net_TlsCertVerify *verify);

/*
 * @brief External log callback function
 *
 * @param level            [in] level, 0/1/2/3 represent debug/info/warn/error
 * @param msg              [in] message, log message with name:code-line-number
 */
typedef void (*Net_LogHandler)(int level, const char *msg);

/*
 * @brief Allocator cache tier policy
 */
typedef enum {
    C_TIER_TIMES = 0, /* tier by times of min-block-size */
    C_TIER_POWER = 1, /* tier by power of min-block-size */
} Net_MemoryAllocatorCacheTierPolicy;

/*
 * @brief Type of allocator
 */
typedef enum {
    C_DYNAMIC_SIZE = 0,            /* allocate dynamic memory size, there is alignment with X KB */
    C_DYNAMIC_SIZE_WITH_CACHE = 1, /* allocator with dynamic memory size, with pre-allocate cache for performance */
} Net_MemoryAllocatorType;

/*
 * @brief Options for Memory Allocator
 */
typedef struct {
    uintptr_t address;                                  /* base address of large range of memory for allocator */
    uint64_t size;                                      /* size of large memory chuck */
    uint32_t minBlockSize;                              /* min size of block, more than 4 KB is required */
    uint32_t bucketCount;                               /* default size of hash bucket */
    uint16_t alignedAddress;                            /* force to align the memory block allocated, 0 means not align
                                                          1 means align */
    uint16_t cacheTierCount;                            /* for DYNAMIC_SIZE_WITH_CACHE only */
    uint16_t cacheBlockCountPerTier;                    /* for DYNAMIC_SIZE_WITH_CACHE only */
    Net_MemoryAllocatorCacheTierPolicy cacheTierPolicy; /* tier policy */
} Net_MemoryAllocatorOptions;

/*
 * @brief memory allocator ptr
 */
typedef uintptr_t Net_MemoryAllocator;

/*
 * @brief Create a memory allocator
 *
 * @param t                [in] type of allocator
 * @param options          [in] options
 * @param allocator        [out] allocator created
 */
int Net_MemoryAllocatorCreate(Net_MemoryAllocatorType t, Net_MemoryAllocatorOptions *options,
    Net_MemoryAllocator *allocator);

/*
 * @brief destroy a memory allocator
 *
 * @param allocator        [in] memory allocator
 *
 * @return 0 if successful
 */
int Net_MemoryAllocatorDestroy(Net_MemoryAllocator allocator);

/*
 * @brief Set the memory region key
 * @param allocator        [in] memory allocator
 *
 * @return 0 if successful
 */
int Net_MemoryAllocatorSetMrKey(Net_MemoryAllocator allocator, uint32_t mrKey);

/*
 * @brief Get the memory offset based on base address
 *
 * @param allocator        [in] memory allocator
 * @param address          [in] memory address
 * @param offset           [out] offset comparing to base address
 *
 * @return 0 if successful
 */
int Net_MemoryAllocatorMemOffset(Net_MemoryAllocator allocator, uintptr_t address, uintptr_t *offset);

/*
 * @brief Get free memory size
 *
 * @param allocator        [in] memory allocator
 *
 * @return 0 if successful
 */
int Net_MemoryAllocatorFreeSize(Net_MemoryAllocator allocator, uintptr_t *size);

/*
 * @brief Allocate memory area
 *
 * @param allocator        [in] memory allocator
 * @param size             [in] size of memory of demand
 * @param address          [out] allocated memory address
 * @param key              [out] allocated memory key
 *
 * @return 0 if successful
 */
int Net_MemoryAllocatorAllocate(Net_MemoryAllocator allocator, uint64_t size, uintptr_t *address, uint32_t *key);

/*
 * @brief Free the address allocated by #Allocate function
 *
 * @param allocator        [in] memory allocator
 * @param address          [in] address to be freed
 *
 * @param 0 if successful
 */
int Net_MemoryAllocatorFree(Net_MemoryAllocator allocator, uintptr_t address);

/*
 * @brief Set external logger function
 *
 * @param h                [in] the log function ptr
 */
void Net_SetExternalLogger(Net_LogHandler h);

/*
 * @brief Check if local host support certain protocol
 *
 * @param t                [in] driver type
 * @param info             [out] driver info
 *
 * @return 1 if supported, 0 if not
 */
int Net_LocalSupport(Net_DriverType t, Net_DeviceInfo *info);
/*
 * @brief Create a driver
 *
 * @param t                [in] type of driver
 * @param name             [in] the name of driver
 * @param startOobSvr      [in] 0 or 1, 1 to start Oob server, 0 don't start Oob server
 * @param driver           [out] created driver address
 *
 * @return 0, if created successfully
 */
int Net_DriverCreate(Net_DriverType t, const char *name, uint8_t startOobSvr, Net_Driver *driver);

/*
 * @brief Set the out of bound ip and port, for endpoint connection
 *
 * @param driver           [in] the address of driver
 * @param ip               [in] the ip for listen or connect
 * @param port             [in] the port for listen or connect
 */
void Net_DriverSetOobIpAndPort(Net_Driver driver, const char *ip, uint16_t port);

/*
 * @brief Get the out of bound ip and port
 *
 * @param driver           [in] the address of driver
 * @param ipArray          [out] oob ip list
 * @param port             [out] oob port list
 * @param length           [out] the length of ipArray and portArray
 */
bool Net_DriverGetOobIpAndPort(Net_Driver driver, char ***ipArray, uint16_t **portArray, int *length);

/*
 * @brief Set oob listener of uds type
 *
 * @param name            [in] name of uds listener
 *
 */
void Net_DriverSetOobUdsName(Net_Driver driver, const char *name);

/*
 * @brief Add multiple oob uds listeners, if there is only one listener just use OobUdsName
 *
 * @param option          [in] option of uds listener option
 *
 */
void Net_DriverAddOobUdsOptions(Net_Driver driver, Net_OobUDSListenerOptions option);

/*
 * @brief Add listen option if to enable multiple listener, duplicated ip and port cannot be added
 *
 * @param driver           [in] the address of driver
 * @param options          [in] the options of the listener
 *
 */
void Net_DriverAddOobOption(Net_Driver driver, Net_DriverOobListenOptions options);

/*
 * @brief Initialize the driver
 *
 * @param driver           [in] the address of driver
 * @param options          [in] options for initialization
 *
 * @return 0 if successful
 */
int Net_DriverInitialize(Net_Driver driver, Net_DriverOptions options);

/*
 * @brief Start the driver, start oob accept thread (server only) and RDMA polling thread
 *
 * @param driver           [in] the address of driver
 *
 * @return 0 if successful
 */
int Net_DriverStart(Net_Driver driver);

/*
 * @brief, Connect to another driver (which is server) and new endpoint will be created if successful
 *
 * There is a retry in it in case of sever is quite busy
 *
 * @param driver           [in] the address of driver
 * @param payloadData      [in] the payloadData, must be ended with \0, i.e. it is a string
 * @param ep               [out] the new endpoint created after connect to server
 * @param flags            [in] flags of ep to be created, NET_C_EP_SELF_POLLING for self polling ep, and
 * NET_C_EP_EVENT_POLLING is the self polling mode
 *
 * @return 0 if successful
 */
int Net_DriverConnect(Net_Driver driver, const char *payloadData, Net_EndPoint *ep, uint32_t flags);

int Net_DriverConnectWithGrpNo(Net_Driver driver, const char *payloadData, Net_EndPoint *ep, uint32_t flags,
    uint8_t serverGrpNo, uint8_t clientGrpNo);

/*
 * @brief Connect to another driver (which is server) and new endpoint will be created if successful
 *
 * There is a retry in it in case of sever is quite busy
 *
 * @param driver           [in] the address of driver
 * @param serverIp         [in] server ip
 * @param serverPort       [in] server listen port
 * @param payloadData      [in] the payloadData, must be ended with \0, i.e. it is a string
 * @param ep               [out] the new endpoint created after connect to server
 * @param flags            [in] flags of ep to be created, NET_C_EP_SELF_POLLING for self polling ep, and
 * NET_C_EP_EVENT_POLLING is the self polling mode
 *
 * @return 0 if successful
 */
int Net_DriverConnectToIpPort(Net_Driver driver, const char *serverIp, uint16_t serverPort, const char *payloadData,
    Net_EndPoint *ep, uint32_t flags);

int Net_DriverConnectIpGrpNo(Net_Driver driver, const char *serverIp, uint16_t serverPort, const char *payloadData,
    Net_EndPoint *ep, uint32_t flags, uint8_t serverGrpNo, uint8_t clientGrpNo);

int Net_DriverConnectWithCtx(Net_Driver driver, const char *serverIp, uint16_t serverPort, const char *payloadData,
    Net_EndPoint *ep, uint32_t flags, uint8_t serverGrpNo, uint8_t clientGrpNo, uint64_t ctx);
/*
 * @brief Stop the driver
 *
 * @param driver           [in] the address of driver
 */
void Net_DriverStop(Net_Driver driver);

/*
 * @brief Un-initialize the driver
 *
 * @param driver           [in] the address of driver
 */
void Net_DriverUnInitialize(Net_Driver driver);

/*
 * @brief Destroy the driver
 *
 * @param driver           [in] the address of driver
 *
 * @return 0 if destroy successful
 */
int Net_DriverDestroy(Net_Driver driver);

/*
 * @brief Enum for callback register [new endpoint connected or endpoint broken]
 */
typedef enum {
    C_EP_NEW = 0,
    C_EP_BROKEN = 1,
} Net_EPHandlerType;

/*
 * @brief Enum for callback register [request received, request posted, read/write done]
 */
typedef enum {
    C_OP_REQUEST_RECEIVED = 0,
    C_OP_REQUEST_POSTED = 1,
    C_OP_READWRITE_DONE = 2,
} Net_OpHandlerType;

/*
 * @brief Register callback function for endpoint
 *
 * @param driver           [in] the address of driver
 * @param t                [in] handle type, could be C_EP_NEW or C_EP_BROKEN
 * @param h                [in] callback function address
 *
 * @return an inner handler address, for un-register in case of memory leak
 */
uintptr_t Net_DriverRegisterEpHandler(Net_Driver driver, Net_EPHandlerType t, Net_EPHandler h, uint64_t usrCtx);

/*
 * @brief Register callback function for endpoint operation
 *
 * @param driver           [in] the address of driver
 * @param t                [in] handle type, could be C_OP_REQUEST_RECEIVED or C_OP_REQUEST_POSTED or
 * C_OP_READWRITE_DONE
 * @param h                [in] callback function address
 *
 * @return an inner handler address, for un-register in case of memory leak
 */
uintptr_t Net_DriverRegisterOpHandler(Net_Driver driver, Net_OpHandlerType t, Net_RequestHandler h, uint64_t usrCtx);

/*
 * @brief Register callback function for worker idle
 *
 * @param driver           [in] the address of driver
 * @param t                [in] handler
 * @param usrCtx           [in] user context, passed to callback function
 *
 * @return an inner handler address, for un-register in case of memory leak
 */
uintptr_t Net_DriverRegisterIdleHandler(Net_Driver driver, Net_IdleHandler h, uint64_t usrCtx);

/*
 * @brief Register callback function for create secure info
 *
 * @param driver           [in] the address of driver
 * @param provider         [in] callback function address
 *
 * @return an inner handler address, for un-register in case of memory leak
 */
uintptr_t Net_DriverRegisterEndpointSecInfoProvider(Net_Driver driver, Net_SecInfoProvider provider);

/*
 * @brief Register callback function for validate secure info from peer
 *
 * @param driver           [in] the address of driver
 * @param validator        [in] callback function address
 *
 * @return an inner handler address, for un-register in case of memory leak
 */
uintptr_t Net_DriverRegisterEndpointSecInfoValidator(Net_Driver driver, Net_SecInfoValidator validator);

/*
 * @brief Register callback function for tls enable
 *
 * @param driver           [in] the address of driver
 * @param certCb           [in] callback to get cert
 * @param priKeyCb         [in] callback to get private key
 * @param caCb             [in] callback to get ca
 *
 * @return an inner handler address, for un-register in case of memory leak
 */
uintptr_t Net_DriverRegisterTLSCb(Net_Driver driver, Net_TlsGetCertCb certCb, Net_TlsGetPrivateKeyCb priKeyCb,
    Net_TlsGetCACb caCb);

/*
 * @brief Un-register callback function for endpoint
 *
 * @param t                [in] handle type, could be C_EP_NEW or C_EP_BROKEN
 * @param handle           [in] callback function address returned when registered
 *
 */
void Net_DriverUnregisterEpHandler(Net_EPHandlerType t, uintptr_t handle);

/*
 * @brief Un-register callback function for endpoint operation
 *
 * @param t                [in] handle type, could be C_OP_REQUEST_RECEIVED or C_OP_REQUEST_POSTED or
 * C_OP_READWRITE_DONE
 * @param handle           [in] callback function address returned when registered
 *
 */
void Net_DriverUnregisterOpHandler(Net_OpHandlerType t, uintptr_t handle);

/*
 * @brief Un-register idle callback
 *
 * @param handle           [in] callback function address returned when registered
 */
void Net_DriverUnregisterIdleHandler(uintptr_t handle);

/*
 * @brief Register a memory region, the memory will be allocated internally
 *
 * @param driver           [in] the address of driver
 * @param size             [in]  size of the memory region
 * @param mr               [out] memory region registered
 *
 * @return 0 successful
 */
int Net_DriverCreateMemoryRegion(Net_Driver driver, uint64_t size, Net_MemoryRegion *mr);

/*
 * @brief Register a memory region, the memory need to be passed in
 *
 * @param driver           [in] the address of driver
 * @param address          [in]  the memory point need to be registered
 * @param size             [in]  size of the memory region
 * @param mr               [out] memory region registered
 *
 * @return 0 successful
 */
int Net_DriverCreateAssignMemoryRegion(Net_Driver driver, uintptr_t address, uint64_t size, Net_MemoryRegion *mr);

/*
 * @brief Unregister the memory region
 *
 * @param driver           [in] the address of driver
 * @param mr               [in] memory region registered
 *
 * @return 0 successful
 */
void Net_DriverDestroyMemoryRegion(Net_Driver driver, Net_MemoryRegion mr);

/*
 * @brief Parse the memory region, get info
 *
 * @param mr               [in] memory region registered
 * @param info             [in] memory region info
 *
 * @return 0 successful
 */
int Net_DriverGetMemoryRegionInfo(Net_MemoryRegion mr, Net_MemoryRegionInfo *info);

/*
 * @brief User can set a relative object address to endpoint
 * this can be used locally only, not send to peer
 *
 * @param ep               [in] address of ep
 * @param ctx              [in] context value to set
 */
void Net_EPSetContext(Net_EndPoint ep, uint64_t ctx);

/*
 * @brief Get the relative object address of the endpoint
 *
 * @param ep               [in] address of ep
 *
 * @return the context set by Net_EPSetContext
 */
uint64_t Net_EPGetContext(Net_EndPoint ep);

#define NET_INVALID_WORKER_INDEX 0xffff
#define NET_INVALID_WORKER_GROUP_INDEX 0xff
/*
 * @brief Get worker index from ep, 0xffff is invalid
 *
 * @param ep               [in] address of ep
 *
 * @return Worker index in the worker group
 */
uint16_t Net_EPGetWorkerIndex(Net_EndPoint ep);

/*
 * @brief Get worker group index from ep, 0xff is invalid
 *
 * @param ep               [in] address of ep
 *
 * @return Group index in the net driver
 */
uint8_t Net_EPGetWorkerGroupIndex(Net_EndPoint ep);

/*
 * @brief Get ep listen port, 0 is invalid
 *
 * @param ep               [in] address of ep
 *
 * @return Listen port of the ep accept from
 */
uint32_t Net_EPGetListenPort(Net_EndPoint ep);

/*
 * @brief Get ep version of peer, the version is transferred when connecting
 *
 * This could be used for version matching for backward compatibility
 *
 * @param ep               [in] address of ep
 *
 * @return Version transferred from peer
 */
uint8_t Net_EPGetVersion(Net_EndPoint ep);

/*
 * @brief Set default timeout
 *
 * 1. timeout = 0: return immediately
 * 2. timeout < 0: never timeout, usually set to -1
 * 3. timeout > 0: second precision timeout.
 */
void Net_EPSetIoTimeout(Net_EndPoint ep, int32_t timeout);

/*
 * @brief Two side RDMA operation, send a data to peer
 *
 * 1) the callback function 'Net_RequestHandler' will be triggered
 * 2) after peer successfully received by RDMA driver, 'Net_RequestHandler' will be trigger as well, i.e. post done
 *
 * @param ep               [in] the endpoint address
 * @param opcode           [in] opcode to peer
 * @param req              [in] request wrappers the data and size
 *
 * @return 0 for successful
 */
int Net_EPPostSend(Net_EndPoint ep, uint16_t opcode, Net_SendRequest *req);

/*
 * @brief Two side RDMA operation, send a data to peer
 *
 * 1) the callback function 'Net_RequestHandler' will be triggered
 * 2) after peer successfully received by RDMA driver, 'Net_RequestHandler' will be trigger as well, i.e. post done
 *
 * @param ep               [in] the endpoint address
 * @param opcode           [in] opcode to peer
 * @param req              [in] request wrappers the data and size
 * @param opInfo           [in] opInfo to peer
 *
 * @return 0 for successful
 */
int Net_EPPostSendWithOpInfo(Net_EndPoint ep, uint16_t opcode, Net_SendRequest *req, Net_OpInfo *opInfo);
/*
 * @brief Two side RDMA operation, send a data to peer
 *
 * 1) the callback function 'Net_RequestHandler' will be triggered
 * 2) after peer successfully received by RDMA driver, 'Net_RequestHandler' will be trigger as well, i.e. post done
 *
 * @param ep               [in] the endpoint address
 * @param opcode           [in] opcode to peer
 * @param req              [in] request wrappers the data and size
 * @param replySeqNo       [in]
 *
 * @return 0 for successful
 */
int Net_EPPostSendWithSeqNO(Net_EndPoint ep, uint16_t opcode, Net_SendRequest *req, uint32_t replySeqNo);

/*
 * @brief Post send a request without opcode and header to peer, peer will be trigger new request callback also
 * without opcode and header, this could be used when you have self define header
 *
 * @param ep               [in] the endpoint address
 * @param req              [in] request information, local address and size is used only, the data is copied, you can
 * free it after called
 * @param seqNo            [in] seq no for peer to reply, must be > 0, peer can get it from context.Header().seqNo,
 * for sync client it will be matching request and response
 *
 * Behavior:
 * 1 For RDMA,
 * case a) if NET_EP_SELF_POLLING is not set, just issue the send request, not wait for sending request finished
 * case b) if NET_EP_SELF_POLLING is set, issue the send request and wait for sending arrived to peer
 *
 * @return 0 if successful
 *
 */
int Net_EPPostSendRaw(Net_EndPoint ep, Net_SendRequest *req, uint32_t seqNo);

/*
 * @brief Post send a request without opcode and header to peer, peer will be trigger new request callback also
 * without opcode and header, this could be used when you have self define header
 *
 * @param request          [in] request information, fill with local different MRs, send to the same remote MR by local
 * MRs sequence, you can free it after called. rKey/rAddress do not need to assign
 * @param seqNo            [in] seq no for peer to reply, must be > 0, peer can get it from context.Header().seqNo,
 * for sync client it will be matching request and response
 *
 * Behavior:
 * 1 For RDMA,
 * case a) if NET_EP_SELF_POLLING is not set, just issue the send request, not wait for sending request finished
 * case b) if NET_EP_SELF_POLLING is set, issue the send request and wait for sending arrived to peer
 *
 * @return 0 if successful
 *
 */
int Net_EPPostSendRawSgl(Net_EndPoint ep, Net_ReadWriteSglRequest *req, uint32_t seqNo);

/*
 * @brief Read RDMA operation, read from peer
 *
 * 1) after peer successfully received by RDMA driver, 'Net_RequestHandler' will be trigger as well, i.e. read done
 *
 * @param ep               [in] the endpoint address
 * @param req              [in] request wrappers the data and size
 *
 * @return 0 for successful
 */
int Net_EPPostRead(Net_EndPoint ep, Net_ReadWriteRequest *req);
int Net_EPPostSglRead(Net_EndPoint ep, Net_ReadWriteSglRequest *req);

/*
 * @brief Write RDMA operation, read from peer
 *
 * 1) after peer successfully received by RDMA driver, 'Net_RequestHandler' will be trigger as well, i.e. read done
 *
 * @param ep               [in] the endpoint address
 * @param req              [in] request wrappers the data and size
 *
 * @return 0 for successful
 */
int Net_EPPostWrite(Net_EndPoint ep, Net_ReadWriteRequest *req);
int Net_EPPostSglWrite(Net_EndPoint ep, Net_ReadWriteSglRequest *req);

/*
 * @brief Wait for send/read/write finish, only for NET_EP_SELF_POLLING is set
 *
 * @param timeout          [in] in second
 * 1. timeout = 0: return immediately
 * 2. timeout < 0: never timeout, usually set to -1
 * 3. timeout > 0: second precision timeout max is 2000s.
 *
 * Behavior:
 * 1 for send, return when request send to peer
 * 2 for read, return when read completion
 * 3 for write, return when write completion
 *
 * @return 0 if successful
 *
 * NN_TIMEOUT if timeout
 */
int Net_EPWaitCompletion(Net_EndPoint ep, int32_t timeout);

/*
 * @brief Get the response for send request reply
 *
 * @param timeout          [in] in second
 * 1. timeout = 0: return immediately
 * 2. timeout < 0: never timeout, usually set to -1
 * 3. timeout > 0: second precision timeout max is 2000s.
 * @param ctx              [out] ctx for response message, ctx cannot be freed by caller
 *
 * @return 0 if successful
 */
int Net_EPReceive(Net_EndPoint ep, int32_t timeout, Net_ResponseContext **ctx);

/*
 * @brief Get the response for send request reply
 *
 * @param timeout          [in] in second
 * 1. timeout = 0: return immediately
 * 2. timeout < 0: never timeout, usually set to -1
 * 3. timeout > 0: second precision timeout max is 2000s.
 * @param ctx              [out] ctx for response message, ctx cannot be freed by caller
 *
 * @return 0 if successful
 */
int Net_EPReceiveRaw(Net_EndPoint ep, int32_t timeout, Net_ResponseContext **ctx);

/*
 * @brief Get the response for send request reply
 *
 * @param timeout          [in] in second
 * 1. timeout = 0: return immediately
 * 2. timeout < 0: never timeout, usually set to -1
 * 3. timeout > 0: second precision timeout max is 2000s.
 * @param ctx              [out] ctx for response message, ctx cannot be freed by caller
 *
 * @return 0 if successful
 */
int Net_EPReceiveRawSgl(Net_EndPoint ep, int32_t timeout, Net_ResponseContext **ctx);

/*
 * @brief Increase the internal reference count, need to call this when forwarding the context to another thread to
 * process
 *
 * @param ep,              [in] the endpoint address
 */
void Net_EPRefer(Net_EndPoint ep);

/*
 * @brief Close endpoint, then will async call broken function
 */
void Net_EPClose(Net_EndPoint ep);

/*
 * @brief Destroy the end point
 *
 * @param ep,              [in] the endpoint address
 */
void Net_EPDestroy(Net_EndPoint ep);

const char *Net_ErrStr(int16_t errCode);

/*
 * @brief Estimated Encrypt length for input raw len
 *
 * @param ep               [in] the endpoint address
 * @param rawLen           [in] raw length before encrypt
 *
 * @return the length after encrypt
 */
uint64_t Net_EstimatedEncryptLen(Net_EndPoint ep, uint64_t rawLen);

/*
 * @brief Encrypt data
 *
 * @param ep               [in] the endpoint address
 * @param rawData          [in] raw data before encrypt
 * @param rawLen           [in] raw data length before encrypt
 * @param cipher           [out] cipher data after encrypt
 * @param cipherLen        [out] cipher data length after encrypt
 *
 * @return 0 if success
 */
int Net_Encrypt(Net_EndPoint ep, const void *rawData, uint64_t rawLen, void *cipher, uint64_t *cipherLen);

/*
 * @brief Estimate Decrypt length
 *
 * @param ep               [in] the endpoint address
 * @param cipherLen        [in] cipher len before decrypt
 *
 * @return the raw length after decrypt
 */
uint64_t Net_EstimatedDecryptLen(Net_EndPoint ep, uint64_t cipherLen);

/*
 * @brief Decrypt data
 *
 * @param ep               [in] the endpoint address
 * @param cipher           [in] cipher data after encrypt
 * @param cipherLen        [in] cipher data length after encrypt
 * @param rawData          [out] raw data before encrypt
 * @param rawLen           [out] raw data length before encrypt
 *
 * @return 0 if success
 */
int Net_Decrypt(Net_EndPoint ep, const void *cipher, uint64_t cipherLen, void *rawData, uint64_t *rawLen);

/*
 * @brief Send shm fds, only shm protocol support
 *
 * @param ep               [in] the endpoint address
 * @param fds              [in] fds to send
 * @param len              [in] fds count to send
 *
 * @return 0 if success
 */
int Net_SendFds(Net_EndPoint ep, int fds[], uint32_t len);

/*
 * @brief Receive shm fds, only shm protocol support
 *
 * @param ep               [in] the endpoint address
 * @param fds              [out] fds to be received
 * @param len              [in] fds count to be received
 * @param timeoutSec       [in] timeout in second for receive. -1 is never timeout
 *
 * @return 0 if success
 */
int Net_ReceiveFds(Net_EndPoint ep, int fds[], uint32_t len, int timeoutSec);

/*
 * @brief Get remote uds ids include pid uid gid, only support in oob server and when oob type is uds
 *
 * @param ep               [in] the endpoint address
 * @param idInfo           [out] remote uds idInfo
 *
 * @return 0 if success
 */
int Net_GetRemoteUdsIdInfo(Net_EndPoint ep, Net_UdsIdInfo *idInfo);
#ifdef __cplusplus
}
#endif

#endif // _OCK_HCOM_12342434341233_H_
