package main

/*
#ifndef BUILD_TESTS
#cgo CFLAGS: -I ../../output/hcom/include/hcom/capi
#cgo LDFLAGS: -L ../../output/hcom/lib -l hcom
#endif
#ifdef BUILD_TESTS
#cgo CFLAGS: -I ../../../../output/hcom/include/hcom/capi
#cgo LDFLAGS: -L ../../../../output/hcom/lib -l hcom
#endif
#include "hcom_service_c.h"
#include "hcom_cgo_c.h"
#include <stdlib.h>
#include <string.h>
*/
import "C"
import (
	"fmt"
	"unsafe"
)

type ServiceOptions struct {
	NetDeviceIpMask string         // ip masks for devices
	EnableTls       uint16         // enable ssl
	tlsMode         TlsMode        // tls mode, default is openssl
	CipherSuite     NetCipherSuite // if tls enabled can set cipher suite, client and server should be same
	/* service setting */
	EnableRndv        uint16 // enable rndv, 1 means enable, 0 means unable
	ConnectWindow     uint16 // server side keep channel alive time in second
	PeriodicThreadNum uint32 // periodic check timeout req thread num, range[1, 4]
	/* worker setting */
	DontStartWorkers   uint16 // start worker or not
	Mode               uint8  // worker polling mode, could busy polling or event polling
	WorkerGroups       string // worker groups, for example 1,3,3
	WorkerGroupsCpuSet string // worker groups cpu set, for example 1-1,2-5,na
	// worker thread priority [-20,20], 20 is the lowest, -20 is the highest, 0 (default) means do not set priority
	WorkerThreadPriority int
	/* connection setting */
	OobType     uint8  // oob type, tcp or UDS, UDS cannot accept remote connection
	LbPolicy    uint8  // select worker load balance policy, default round-robin
	MaxTypeSize uint16 // service start() will check handlers with max size
	Magic       uint16 // magic number for c/s connect validation
	Version     uint8  // program version used by connect validation
	/* heart beat setting */
	HeartBeatIdleTime      uint16 // heart beat idle time, in seconds
	HeartBeatProbeTimes    uint16 // heart beat probe times
	HeartBeatProbeInterval uint16 // heart beat probe interval, in seconds
	/* options for rdma protocol only */
	MrSendReceiveSegCount   uint32 // memory region segment count for two side operation
	MrSendReceiveSegSize    uint32 // data size of memory region segment
	CompletionQueueDepth    uint16 // completion queue size of rdma
	MaxPostSendCountPerQP   uint16 // max number request could issue
	PrePostReceiveSizePerQP uint16 // pre post receive of qp
	PollingBatchSize        uint16 // polling batch size for worker
	EventPollingTimeout     uint16 // event polling timeout in us
	QpSendQueueSize         uint32 // max send working request of qp for rdma
	QpReceiveQueueSize      uint32 // max receive working request of qp for rdma
}

type ServiceConnectOpt struct {
	Index       uint16 /* connect type index for different op handlers */
	EpSize      uint8  /* assign ep size in channel */
	CbType      uint8  /* channel callback mode */
	ClientGrpNo uint8  /* indicates which server worker group to connect to */
	ServerGrpNo uint8  /* indicates which client worker group to connect to */
	Flags       uint32 /* for server default 0, for client sync ep :SER_C_CHANNEL_SELF_POLLING SER_C_CHANNEL_EVENT_POLLING */
}

type ServiceOpInfo struct {
	OpCode    uint16 // operation code, 0~999
	Timeout   int16  // timeout in seconds
	Flags     uint8  // flags in user case
	ErrorCode int16  // error code
}

type ServiceMemoryRegionInfo struct {
	LAddress uintptr // local memory region address
	LKey     uint32  // local memory region key
	Size     uint32  // data size
}

type ServiceRequest struct {
	LAddress uintptr // local buffer address
	RAddress uintptr // remote buffer address
	LKey     uint32  // local memory region key
	RKey     uint32  // remote memory region key
	Size     uint32  // buffer size
}

type MemoryAllocatorOptions struct {
	Address                Memory                   /* base address of large range of memory for allocator */
	Size                   uint64                   /* size of large memory chuck */
	MinBlockSize           uint32                   /* min size of block, more than 4 KB is required */
	BucketCount            uint32                   /* default size of hash bucket */
	AlignedAddress         uint16                   /* force to align the memory block allocated, 0 means not align, 1 means align */
	CacheTierCount         uint16                   /* for DYNAMIC_SIZE_WITH_CACHE only */
	CacheBlockCountPerTier uint16                   /* for DYNAMIC_SIZE_WITH_CACHE only */
	CacheTierPolicy        AllocatorCacheTierPolicy /* tier policy */
}

type ChannelCallback struct {
	Cb  ChannelCallbackFunc // User callback function, the method signature is func(arg unsafe.Pointer, context ServiceContext)
	Arg unsafe.Pointer      // Argument of callback
}

type ServiceMsg struct {
	Data unsafe.Pointer // User callback function
	Size uint32         // Argument of callback
}

// ServiceCreate
// @brief Create a service
//
// @param serviceType      [in] type of service
// @param name             [in] the name of service
// @param startOobSvr      [in] 0 or 1, 1 to start Oob server, 0 don't start Oob server
//
// @return 0, if created successfully
// @return created service address
func ServiceCreate(serviceType ServiceType, name string, startOobSvr uint8) (int, NetService) {
	if serviceType != 0 && serviceType != 1 {
		fmt.Println("invalid serviceType")
		return -1, 0
	}
	service := C.Net_Service(0)
	res := C.Service_Create(C.Service_Type(serviceType), C.CString(name), C.uint8_t(startOobSvr), &service)
	return int(res), NetService(service)
}

// ServiceStart
// @brief Start the driver, start oob accept thread (server only) and RDMA polling thread
//
// @param service          [in] the address of service
// @param options          [in] the options of the service
//
// @return 0, if successful
func ServiceStart(service NetService, options ServiceOptions) int {
	var cOptions C.Service_Options
	C.Cgo_StringCopy(&cOptions.netDeviceIpMask[0], NO256, C.CString(options.NetDeviceIpMask))
	cOptions.enableTls = C.uint16_t(options.EnableTls)
	cOptions.enableRndv = C.uint16_t(options.EnableRndv)
	cOptions.connectWindow = C.uint16_t(options.ConnectWindow)
	cOptions.periodicThreadNum = C.uint32_t(options.PeriodicThreadNum)
	cOptions.dontStartWorkers = C.uint16_t(options.DontStartWorkers)
	cOptions.mode = C.Service_WorkingMode(options.Mode)
	C.Cgo_StringCopy(&cOptions.workerGroups[0], NO64, C.CString(options.WorkerGroups))
	C.Cgo_StringCopy(&cOptions.workerGroupsCpuSet[0], NO128, C.CString(options.WorkerGroupsCpuSet))
	cOptions.workerThreadPriority = C.int(options.WorkerThreadPriority)
	cOptions.oobType = C.Service_OobType(options.OobType)
	cOptions.lbPolicy = C.Service_LBPolicy(options.LbPolicy)
	cOptions.maxTypeSize = C.uint16_t(options.MaxTypeSize)
	cOptions.magic = C.uint16_t(options.Magic)
	cOptions.version = C.uint8_t(options.Version)
	cOptions.heartBeatIdleTime = C.uint16_t(options.HeartBeatIdleTime)
	cOptions.heartBeatProbeTimes = C.uint16_t(options.HeartBeatProbeTimes)
	cOptions.heartBeatProbeInterval = C.uint16_t(options.HeartBeatProbeInterval)
	cOptions.mrSendReceiveSegCount = C.uint32_t(options.MrSendReceiveSegCount)
	cOptions.mrSendReceiveSegSize = C.uint32_t(options.MrSendReceiveSegSize)
	cOptions.completionQueueDepth = C.uint16_t(options.CompletionQueueDepth)
	cOptions.maxPostSendCountPerQP = C.uint16_t(options.MaxPostSendCountPerQP)
	cOptions.prePostReceiveSizePerQP = C.uint16_t(options.PrePostReceiveSizePerQP)
	cOptions.pollingBatchSize = C.uint16_t(options.PollingBatchSize)
	cOptions.eventPollingTimeout = C.uint16_t(options.EventPollingTimeout)
	cOptions.qpSendQueueSize = C.uint32_t(options.QpSendQueueSize)
	cOptions.qpReceiveQueueSize = C.uint32_t(options.QpReceiveQueueSize)
	res := C.Service_Start(C.Net_Service(service), cOptions)
	return int(res)
}

// ServiceStop
// @brief Stop the service
//
// @param service          [in] the address of service
func ServiceStop(service NetService) {
	C.Service_Stop(C.Net_Service(service))
}

// ServiceDestroy
// @brief Destroy the service
//
// @param service          [in] the address of service
//
// @return 0, if destroy successful
func ServiceDestroy(service NetService) int {
	return int(C.Service_Destroy(C.Net_Service(service)))
}

// ServiceConnect
// @brief Destroy the service
//
// @param service          [in] the address of service
// @param oobIpOrName      [in] oob ip or name to connect, set ip for tcp and name for uds
// @param oobPort          [in] only need to set when tcp oob
// @param payload          [in] payload transferred to peer, will be got EP Connected NetCallback at server
// @param options          [in] variable argument for specific use
//
// @return 0 successful
// @return connected channel
func ServiceConnect(service NetService, oobIpOrName string, oobPort uint16, payload string, opt ServiceConnectOpt) (int, NetChannel) {
	channel := C.Net_Channel(0)
	var cOptions C.Service_ConnectOpt
	cOptions.index = C.uint16_t(opt.Index)
	cOptions.epSize = C.uint8_t(opt.EpSize)
	cOptions.cbType = C.Channel_CBType(opt.CbType)
	cOptions.clientGrpNo = C.uint8_t(opt.ClientGrpNo)
	cOptions.serverGrpNo = C.uint8_t(opt.ServerGrpNo)
	cOptions.flags = C.uint32_t(opt.Flags)
	res := C.Service_Connect(C.Net_Service(service), C.CString(oobIpOrName), C.uint16_t(oobPort), C.CString(payload), &channel, &cOptions)
	return int(res), NetChannel(channel)
}

// ChannelClose
// @brief Destroy the service
//
// @param channel          [in] connected channel
func ChannelClose(channel NetChannel) {
	C.Channel_Close(C.Net_Channel(channel))
}

// RegisterNewEndPointHandler
// @brief Register new endpoint callback function for channel
//
// @param service          [in] the address of service
// @param handler          [in] callback function address, the method signature is func(channel uintptr, usrCtx uint64, payload *C.char) int
// @param usrCtx           [in] ctx that could be passed to the handler
//
// @return an inner handler address, for un-register in case of memory leak. Return 0 means failed
func RegisterNewEndPointHandler(service NetService, handler NewEndPointHandler, userCtx uint64) uintptr {
	res := C.Service_RegisterChannelHandler(C.Net_Service(service), 0, C.Service_ChannelHandler(handler), C.Service_ChannelPolicy(0), C.uint64_t(userCtx))
	return uintptr(res)
}

// RegisterBrokenEndPointHandler
// @brief Register endpoint broken callback function for channel
//
// @param service          [in] the address of service
// @param handler          [in] callback function address, the method signature is func(channel uintptr, usrCtx uint64, payload *C.char) int
// @param policy           [in] ep broken policy
// @param usrCtx           [in] ctx that could be passed to the handler
//
// @return an inner handler address, for un-register in case of memory leak. Return 0 means failed
func RegisterBrokenEndPointHandler(service NetService, handler BrokenEndPointHandler, policy ServiceChannelPolicy, userCtx uint64) uintptr {
	return uintptr(C.Service_RegisterChannelHandler(C.Net_Service(service), 1, C.Service_ChannelHandler(handler), C.Service_ChannelPolicy(policy), C.uint64_t(userCtx)))
}

// RegisterRequestReceivedHandler
// @brief Register request received callback function for channel
//
// @param service          [in] the address of service
// @param handler          [in] callback function address, the method signature is func(ctx uintptr, usrCtx uint64) int
// @param index            [in] channel type index, range [0, 15]
// @param usrCtx           [in] ctx that could be passed to the handler
//
// @return an inner handler address, for un-register in case of memory leak. Return 0 means failed
func RegisterRequestReceivedHandler(service NetService, handler RequestReceivedHandler, index uint16, userCtx uint64) uintptr {
	return uintptr(C.Service_RegisterOpHandler(C.Net_Service(service), C.uint16_t(index), 0, C.Service_ChannelHandler(handler), C.uint64_t(userCtx)))
}

// RegisterRequestPostHandler
// @brief Register request post callback function for channel
//
// @param service          [in] the address of service
// @param handler          [in] callback function address, the method signature is func(ctx uintptr, usrCtx uint64) int
// @param index            [in] channel type index, range [0, 15]
// @param usrCtx           [in] ctx that could be passed to the handler
//
// @return an inner handler address, for un-register in case of memory leak. Return 0 means failed
func RegisterRequestPostHandler(service NetService, handler RequestPostHandler, index uint16, userCtx uint64) uintptr {
	return uintptr(C.Service_RegisterOpHandler(C.Net_Service(service), C.uint16_t(index), 1, C.Service_ChannelHandler(handler), C.uint64_t(userCtx)))
}

// RegisterOneSideDoneHandler
// @brief Register read write callback function for channel
//
// @param service          [in] the address of service
// @param handler          [in] callback function address, the method signature is func(ctx uintptr, usrCtx uint64) int
// @param index            [in] channel type index, range [0, 15]
// @param usrCtx           [in] ctx that could be passed to the handler
//
// @return an inner handler address, for un-register in case of memory leak. Return 0 means failed
func RegisterOneSideDoneHandler(service NetService, handler OneSideDoneHandler, index uint16, userCtx uint64) uintptr {
	return uintptr(C.Service_RegisterOpHandler(C.Net_Service(service), C.uint16_t(index), 2, C.Service_ChannelHandler(handler), C.uint64_t(userCtx)))
}

// RegisterTLSCb
// @brief Register tls callback function for channel
//
// @param service          [in] the address of service
// @param getCertCb        [in] callback function address, the method signature is func(*C.char, **C.char) int
// @param getPrivateKeyCb  [in] callback function address, the method signature is func(*C.char, **C.char, **C.char, *C.Net_TlsKeyPassErase) int
// @param getCACb          [in] callback function address, the method signature is func( *C.char, **C.char, **C.char, *int32, *C.Net_TlsCertVerify) int
//
// @return an inner handler address, for un-register in case of memory leak. Return 0 means failed
func RegisterTLSCb(service NetService, getCertCb TlsGetCertCb, getPrivateKeyCb TlsGetPrivateKeyCb, getCACb TlsGetCACb) uintptr {
	return uintptr(C.Service_RegisterTLSCb(C.Net_Service(service), C.Net_TlsGetCertCb(getCertCb), C.Net_TlsGetPrivateKeyCb(getPrivateKeyCb), C.Net_TlsGetCACb(getCACb)))
}

// SetOobIpAndPort
// @brief Set out of bound ip and port
//
// @param service          [in] the address of service
// @param ip               [in] ip address
// @param port             [in] port
func SetOobIpAndPort(service NetService, ip string, port uint16) {
	C.Service_SetOobIpAndPort(C.Net_Service(service), C.CString(ip), C.uint16_t(port))
}

// PrepareMsg
// @brief Prepare the msg for the communication
//
// @param msg              [in] the address of msg
// @param size             [in] size of the msg
//
// @return the prepared msg
func PrepareMsg(msg unsafe.Pointer, size uint32) ServiceMsg {
	var serviceMsg ServiceMsg
	cMsg := C.Cgo_AllocMem(C.uint32_t(size))
	if cMsg == nil {
		return serviceMsg
	}
	if C.Cgo_MemoryCopy(cMsg, msg, C.uint32_t(size), C.uint32_t(size)) != 0 {
		return serviceMsg
	}
	serviceMsg.Data = unsafe.Pointer(cMsg)
	serviceMsg.Size = size
	return serviceMsg
}

// MallocMemory
// @brief allocate memory in c
//
// @param size             [in] the size of memory
//
// @return the allocated memory
func MallocMemory(size uint32) MemoryPointer {
	cMsg := C.Cgo_AllocMem(C.uint32_t(size))
	return MemoryPointer(cMsg)
}

// ReleaseMsg
// @brief release the msg for the communication
//
// @param msg              [in] the prepared msg
func ReleaseMsg(msg ServiceMsg) {
	C.Cgo_FreeMem(msg.Data)
}

// ReleaseData
// @brief release the data for the communication
//
// @param data              [in] the data from
func ReleaseData(data MemoryPointer) {
	C.Cgo_FreeMem(unsafe.Pointer(data))
}

// ServiceGetResult
// @brief Get the result
//
// @param ctx              [in] the ctx param in the callback function
//
// @return 0 means OK, other value means failed.
// @return the error code, 0 means OK, other value indicates the error
func ServiceGetResult(ctx ServiceContext) (int, int) {
	result := C.int(0)
	res := C.Service_GetResult(C.Service_Context(ctx), &result)
	return int(res), int(result)
}

// ServiceGetChannel
// @brief Get the channel
//
// @param ctx              [in] the ctx param in the callback function
//
// @return 0 means OK, other value means failed.
// @return the channel
func ServiceGetChannel(ctx ServiceContext) (int, NetChannel) {
	channel := C.Net_Channel(0)
	res := C.Service_GetChannel(C.Service_Context(ctx), &channel)
	return int(res), NetChannel(channel)
}

// ServiceGetRndvChannel
// @brief Get the channel
//
// @param ctx              [in] the ctx param in the callback function
//
// @return 0 means OK, other value means failed.
// @return the channel
func ServiceGetRndvChannel(ctx ServiceRndvContext) (int, NetChannel) {
	channel := C.Net_Channel(0)
	res := C.Service_GetRndvChannel(C.Service_RndvContext(ctx), &channel)
	return int(res), NetChannel(channel)
}

// ServiceGetMessageData
// @brief Get message data received which valid in callback lifetime
//
// @param ctx              [in] the ctx param in the callback function
//
// @return valid address if success, NULL if failed
func ServiceGetMessageData(ctx ServiceContext) MessageData {
	return MessageData(C.Service_GetMessageData(C.Service_Context(ctx)))
}

// ServiceGetMessageDataLen
// @brief Get message data length received which valid in callback lifetime
//
// @param ctx              [in] the ctx param in the callback function
//
// @return valid length if success, 0 if failed
func ServiceGetMessageDataLen(ctx ServiceContext) uint32 {
	return uint32(C.Service_GetMessageDataLen(C.Service_Context(ctx)))
}

// ServiceGetRndvMessage
// @brief Get message data length received which valid in callback lifetime
//
// @param ctx              [in] the ctx param in the callback function
//
// @return valid length if success, 0 if failed
func ServiceGetRndvMessage(ctx ServiceRndvContext) (int, MessageData, uint32) {
	var cServiceMsg C.Service_Message
	res := C.Cgo_GetRndvMessage(C.Service_RndvContext(ctx), &cServiceMsg)
	if res != 0 {
		return int(res), nil, 0
	}

	return int(res), MessageData(cServiceMsg.data), uint32(cServiceMsg.size)
}

// ServiceGetContextType
// @brief Get the context type
//
// @param ctx              [in] the ctx param in the callback function
//
// @return 0 means OK, other value means failed
// @return if ok, the ctx type
func ServiceGetContextType(ctx ServiceContext) (int, ServiceContextType) {
	ctxType := C.Service_ContextType(0)
	res := C.Service_GetContextType(C.Service_Context(ctx), &ctxType)
	return int(res), ServiceContextType(ctxType)
}

// ServiceGetRndvCtxType
// @brief Get the rndv context type
//
// @param ctx              [in] the ctx param in the callback function
//
// @return 0 means OK, other value means failed
// @return the ctx type
func ServiceGetRndvCtxType(ctx ServiceRndvContext) (int, ServiceRndvType) {
	ctxType := C.Service_RndvType(0)
	res := C.Service_GetRndvCtxType(C.Service_RndvContext(ctx), &ctxType)
	return int(res), ServiceRndvType(ctxType)
}

// ServiceGetOpInfo
// @brief Get op info by user input
//
// @param ctx              [in] the ctx param in the callback function
//
// @return 0 means OK, other value means failed
// @return if ok, the op info
func ServiceGetOpInfo(ctx ServiceContext) (int, ServiceOpInfo) {
	var cOpInfo C.Service_OpInfo
	res := C.Service_GetOpInfo(C.Service_Context(ctx), &cOpInfo)
	var opInfo ServiceOpInfo
	opInfo.OpCode = uint16(cOpInfo.opCode)
	opInfo.Timeout = int16(cOpInfo.timeout)
	opInfo.Flags = uint8(cOpInfo.flags)
	opInfo.ErrorCode = int16(cOpInfo.errorCode)
	return int(res), opInfo
}

// ServiceGetRndvOpInfo
// @brief Get op info by user input
//
// @param ctx              [in] the ctx param in the callback function
//
// @return 0 means OK, other value means failed
// @return the op info
func ServiceGetRndvOpInfo(ctx ServiceRndvContext) (int, ServiceOpInfo) {
	var cOpInfo C.Service_OpInfo
	res := C.Service_GetRndvOpInfo(C.Service_RndvContext(ctx), &cOpInfo)
	var opInfo ServiceOpInfo
	opInfo.OpCode = uint16(cOpInfo.opCode)
	opInfo.Timeout = int16(cOpInfo.timeout)
	opInfo.Flags = uint8(cOpInfo.flags)
	opInfo.ErrorCode = int16(cOpInfo.errorCode)
	return int(res), opInfo
}

// ServiceRndvFreeContext
// @brief Free rndv context
//
// @param ctx              [in] the ctx param in the callback function
//
// @return 0 means OK, other value means failed
func ServiceRndvFreeContext(ctx ServiceRndvContext) int {
	return int(C.Service_RndvFreeContext(C.Service_RndvContext(ctx)))
}

// ServiceGetRspCtx
// @brief Get response context for send rsp message
//
// @param ctx              [in] the ctx param in the callback function
//
// @return 0 means OK, other value means failed
// @return if ok, the response context
func ServiceGetRspCtx(ctx ServiceContext) (int, ServiceRspCtx) {
	rsp := C.Service_RspCtx(0)
	res := C.Service_GetRspCtx(C.Service_Context(ctx), &rsp)
	return int(res), ServiceRspCtx(rsp)
}

// ChannelPostResponse
// @brief Post send response message with op info to peer, and without waiting for response. Peer will be trigger
// new request NetCallback also with op info
//
// @param channel       [in] net channel
// @param rspCtx        [in] get rspCtx from ServiceGetRspCtx
// @param opInfo        [in] get opInfo from ServiceGetOpInfo
// @param message       [in] user single common message
// @param cb            [in] NetCallback for async send to peer. If cb is nullptr, the interface will block until
// complete. And it means if set cb is nullptr in worker callback context, the interface will block worker thread.
//
// @return 0, if cb is null, it means the operation is cb successfully. Otherwise, it means the operation
// is in progress. return ERROR, cb will not be invoked.
func ChannelPostResponse(channel NetChannel, rspCtx ServiceRspCtx, opInfo ServiceOpInfo, msg ServiceMsg, callback ChannelCallback) int {
	var cOpInfo C.Service_OpInfo
	cOpInfo.opCode = C.uint16_t(opInfo.OpCode)
	cOpInfo.timeout = C.int16_t(opInfo.Timeout)
	cOpInfo.flags = C.uint8_t(opInfo.Flags)
	cOpInfo.errorCode = C.int16_t(opInfo.ErrorCode)
	var cMsg C.Service_Message
	cMsg.data = msg.Data
	cMsg.size = C.uint32_t(msg.Size)
	var res C.int
	if callback.Cb != nil {
		var cb C.Channel_Callback
		cb.cb = C.Channel_CallbackFunc(callback.Cb)
		cb.arg = callback.Arg
		res = C.Channel_PostResponse(C.Net_Channel(channel), C.Service_RspCtx(rspCtx), &cOpInfo, &cMsg, &cb)
	} else {
		res = C.Channel_PostResponse(C.Net_Channel(channel), C.Service_RspCtx(rspCtx), &cOpInfo, &cMsg, nil)
	}
	return int(res)
}

// ChannelSyncCall
// @brief Post send a request with op info to peer, and waiting for response. Peer will be trigger new
// request NetCallback also with op info
//
// @param channel       [in] net channel
// @param reqInfo       [in] peer can get opInfo from ServiceGetOpInfo
// @param req           [in] prepared msg send to remote.
//
// @return 0 means OK, other value means failed.
// @return peer response info
// @return receive remote response message
func ChannelSyncCall(channel NetChannel, reqInfo ServiceOpInfo, req ServiceMsg) (int, ServiceOpInfo, ServiceMsg) {
	var cMsg C.Service_Message
	cMsg.data = req.Data
	cMsg.size = C.uint32_t(req.Size)
	var cReqInfo C.Service_OpInfo
	cReqInfo.opCode = C.uint16_t(reqInfo.OpCode)
	cReqInfo.timeout = C.int16_t(reqInfo.Timeout)
	cReqInfo.flags = C.uint8_t(reqInfo.Flags)
	cReqInfo.errorCode = C.int16_t(reqInfo.ErrorCode)
	var cRsp C.Service_Message
	var cRspInfo C.Service_OpInfo
	res := C.Channel_SyncCall(C.Net_Channel(channel), &cReqInfo, &cMsg, &cRspInfo, &cRsp)
	var rspInfo ServiceOpInfo
	rspInfo.OpCode = uint16(cRspInfo.opCode)
	rspInfo.Timeout = int16(cRspInfo.timeout)
	rspInfo.Flags = uint8(cRspInfo.flags)
	rspInfo.ErrorCode = int16(cRspInfo.errorCode)
	var rsp ServiceMsg
	rsp.Data = unsafe.Pointer(cRsp.data)
	rsp.Size = uint32(cRsp.size)
	return int(res), rspInfo, rsp
}

// ChannelAsyncCall
// @brief Post send request message with op info to peer, and without waiting for response. Peer will be trigger
// new request NetCallback also with op info
//
// @param channel       [in] net channel
// @param opInfo        [in] peer can get opInfo from ServiceGetOpInfo
// @param req           [in] send to remote message.
// @param callback      [in] receive message will store in NetServiceContext::Message() in callback lifetime. If
// user want to use Message() in other thread, need to prepare buff and copy data out.
//
// @return 0 means OK, other value means failed.
func ChannelAsyncCall(channel NetChannel, opInfo ServiceOpInfo, req ServiceMsg, callback ChannelCallback) int {
	if callback.Cb == nil {
		fmt.Println("callback could not be null in async call")
		return -1
	}
	var cMsg C.Service_Message
	cMsg.data = req.Data
	cMsg.size = C.uint32_t(req.Size)
	var cOpInfo C.Service_OpInfo
	cOpInfo.opCode = C.uint16_t(opInfo.OpCode)
	cOpInfo.timeout = C.int16_t(opInfo.Timeout)
	cOpInfo.flags = C.uint8_t(opInfo.Flags)
	cOpInfo.errorCode = C.int16_t(opInfo.ErrorCode)
	var cb C.Channel_Callback
	cb.cb = C.Channel_CallbackFunc(callback.Cb)
	cb.arg = callback.Arg
	res := C.Channel_AsyncCall(C.Net_Channel(channel), &cOpInfo, &cMsg, &cb)
	return int(res)
}

// ChannelGetPeerIpPort
// @brief Get peer ip and port
//
// @param channel       [in] net channel
//
// @return 0 means OK, other value means failed.
// @return ip
// @return port
func ChannelGetPeerIpPort(channel NetChannel) (int, string, uint16) {
	var cIp *C.char
	var cPort C.uint16_t
	res := C.Channel_GetPeerIpPort(C.Net_Channel(channel), &cIp, &cPort)
	defer C.free(unsafe.Pointer(cIp))
	if res == 0 {
		return int(res), C.GoString(cIp), uint16(cPort)
	} else {
		return int(res), "", 0
	}
}

// ServiceSetExternalLogger
// @brief Set external logger function
//
// @param h                [in] the log function ptr, the method signature is func(level int, msg *C.char)
//
// @return 0 means OK, other value means failed
// @return the channel
func ServiceSetExternalLogger(logHandler ServiceLogHandler) {
	C.Service_SetExternalLogger(C.Service_LogHandler(logHandler))
}

// RegisterAllocateHandler
// @brief Register callback function for rndv memory allocate
//
// @param service          [in] the address of service
// @param handler          [in] callback function address, the method signature is func(size uint64, address *uintptr, key *uint32) int
//
// @return an inner handler address, for un-register in case of memory leak. Return 0 means failed
func RegisterAllocateHandler(service NetService, handler RndvMemAllocateHandler) uintptr {
	return uintptr(C.Service_RegisterAllocateHandler(C.Net_Service(service), C.Service_RndvMemAllocate(handler)))
}

// RegisterFreeHandler
// @brief Register callback function for rndv memory free
//
// @param service          [in] the address of service
// @param handler          [in] callback function address, the method signature is func(address uintptr) int
//
// @return an inner handler address, for un-register in case of memory leak. Return 0 means failed
func RegisterFreeHandler(service NetService, handler RndvMemFreeHandler) uintptr {
	return uintptr(C.Service_RegisterFreeHandler(C.Net_Service(service), C.Service_RndvMemFree(handler)))
}

// RegisterRndvOpHandler
// @brief Register callback function for rndv op handler
//
// @param service          [in] the address of service
// @param handler          [in] callback function address, the method signature is func(ctx NetServiceRndvContext) int
//
// @return an inner handler address, for un-register in case of memory leak. Return 0 means failed
func RegisterRndvOpHandler(service NetService, handler RndvHandler) uintptr {
	return uintptr(C.Service_RegisterRndvOpHandler(C.Net_Service(service), C.Service_RndvHandler(handler)))
}

// RndvReply
// @brief Reply message with op info to peer rndv msg
//
// @param context       [in] rndv context
// @param opInfo        [in] peer can get opInfo from Service_GetOpInfo
// @param msg           [in] user single common message
// @param callback      [in] NetCallback for async send to peer. If cb is nullptr, the interface will block until
// complete. And it means if set cb is nullptr in worker callback context, the interface will block worker thread.
//
// @return 0 if cb is null, it means the operation is cb successfully. Otherwise, it means the operation
// is in progress. return ERROR, cb will not be invoked.
func RndvReply(context ServiceRndvContext, opInfo ServiceOpInfo, msg ServiceMsg, callback ChannelCallback) int {
	var cOpInfo C.Service_OpInfo
	cOpInfo.opCode = C.uint16_t(opInfo.OpCode)
	cOpInfo.timeout = C.int16_t(opInfo.Timeout)
	cOpInfo.flags = C.uint8_t(opInfo.Flags)
	cOpInfo.errorCode = C.int16_t(opInfo.ErrorCode)
	var cMsg C.Service_Message
	cMsg.data = msg.Data
	cMsg.size = C.uint32_t(msg.Size)
	var res C.int
	if callback.Cb != nil {
		var cb C.Channel_Callback
		cb.cb = C.Channel_CallbackFunc(callback.Cb)
		cb.arg = callback.Arg
		res = C.Service_RndvReply(C.Service_RndvContext(context), &cOpInfo, &cMsg, &cb)
	} else {
		res = C.Service_RndvReply(C.Service_RndvContext(context), &cOpInfo, &cMsg, nil)
	}
	return int(res)
}

// AlignMemory
// @brief memalign memory
//
// @param alignment        [in] alignment size
// @param size             [in] memory size
//
// @return the aligned memory
func AlignMemory(alignment uint32, size uint64) Memory {
	cMem := C.Cgo_AlignMem(C.uint32_t(alignment), C.uint64_t(size))
	return Memory(cMem)
}

// MemoryAllocatorCreate
// @brief Create a memory allocator
//
// @param allocatorType    [in] type of allocator
// @param options          [in] options
//
// @return 0, if created successfully
// @return created allocator
func MemoryAllocatorCreate(allocatorType AllocatorType, options MemoryAllocatorOptions) (int, MemoryAllocator) {
	var cOptions C.Net_MemoryAllocatorOptions
	cOptions.address = C.uintptr_t(options.Address)
	cOptions.size = C.uint64_t(options.Size)
	cOptions.minBlockSize = C.uint32_t(options.MinBlockSize)
	cOptions.bucketCount = C.uint32_t(options.BucketCount)
	cOptions.alignedAddress = C.uint16_t(options.AlignedAddress)
	cOptions.cacheTierCount = C.uint16_t(options.CacheTierCount)
	cOptions.cacheBlockCountPerTier = C.uint16_t(options.CacheBlockCountPerTier)
	cOptions.cacheTierPolicy = C.Net_MemoryAllocatorCacheTierPolicy(options.CacheTierPolicy)
	allocator := C.Net_MemoryAllocator(0)
	res := C.Net_MemoryAllocatorCreate(C.Net_MemoryAllocatorType(allocatorType), &cOptions, &allocator)
	return int(res), MemoryAllocator(allocator)
}

// MemoryAllocatorDestroy
// @brief Destroy a memory allocator
//
// @param allocator        [in] memory allocator
//
// @return 0, if successful
func MemoryAllocatorDestroy(allocator MemoryAllocator) int {
	res := C.Net_MemoryAllocatorDestroy(C.Net_MemoryAllocator(allocator))
	return int(res)
}

// MemoryAllocatorSetMrKey
// @brief Set the memory region key
//
// @param allocator        [in] memory allocator
// @param mrKey			   [in] memory region key
//
// @return 0, if successful
func MemoryAllocatorSetMrKey(allocator MemoryAllocator, mrKey uint32) int {
	res := C.Net_MemoryAllocatorSetMrKey(C.Net_MemoryAllocator(allocator), C.uint32_t(mrKey))
	return int(res)
}

// MemoryAllocatorAllocate
// @brief Allocate memory area
//
// @param allocator        [in] memory allocator
// @param size             [in] size of memory of demand
//
// @return 0, if created successfully
// @return allocated memory address
// @return allocated memory key
func MemoryAllocatorAllocate(allocator MemoryAllocator, size uint64) (int, Memory, uint32) {
	var address C.uintptr_t
	var key C.uint32_t
	res := C.Net_MemoryAllocatorAllocate(C.Net_MemoryAllocator(allocator), C.uint64_t(size), &address, &key)
	return int(res), Memory(address), uint32(key)
}

// MemoryAllocatorFree
// @brief Free the address allocated by #Allocate function
//
// @param allocator        [in] memory allocator
// @param address          [in] address to be freed
//
// @return 0, if successful
func MemoryAllocatorFree(allocator MemoryAllocator, address Memory) int {
	return int(C.Net_MemoryAllocatorFree(C.Net_MemoryAllocator(allocator), C.uintptr_t(address)))
}

// ChannelSyncRndvCall
// @brief Post send a register MR with op info to peer, and waiting for response. Peer will be trigger new
//
//	request NetCallback also with op info. Poll mode:worker poll(default mode, support sync/async with sem), self
//	poll(support sync only)
//
// @param channel       [in] net channel
// @param reqInfo       [in] peer can get opInfo from ServiceGetOpInfo
// @param req           [in] prepared msg send to remote
//
// @return 0 means OK, other value means failed
// @return peer response info
// @return receive remote response message
func ChannelSyncRndvCall(channel NetChannel, reqInfo ServiceOpInfo, req ServiceRequest) (int, ServiceOpInfo, ServiceMsg) {
	var cReq C.Service_Request
	cReq.lAddress = C.uintptr_t(req.LAddress)
	cReq.rAddress = C.uintptr_t(req.RAddress)
	cReq.lKey = C.uint32_t(req.LKey)
	cReq.rKey = C.uint32_t(req.RKey)
	cReq.size = C.uint32_t(req.Size)
	var cReqInfo C.Service_OpInfo
	cReqInfo.opCode = C.uint16_t(reqInfo.OpCode)
	cReqInfo.timeout = C.int16_t(reqInfo.Timeout)
	cReqInfo.flags = C.uint8_t(reqInfo.Flags)
	cReqInfo.errorCode = C.int16_t(reqInfo.ErrorCode)
	var cRsp C.Service_Message
	var cRspInfo C.Service_OpInfo
	res := C.Channel_SyncRndvCall(C.Net_Channel(channel), &cReqInfo, &cReq, &cRspInfo, &cRsp)
	var rspInfo ServiceOpInfo
	rspInfo.OpCode = uint16(cRspInfo.opCode)
	rspInfo.Timeout = int16(cRspInfo.timeout)
	rspInfo.Flags = uint8(cRspInfo.flags)
	rspInfo.ErrorCode = int16(cRspInfo.errorCode)
	var rsp ServiceMsg
	rsp.Data = unsafe.Pointer(cRsp.data)
	rsp.Size = uint32(cRsp.size)
	return int(res), rspInfo, rsp
}

// ChannelAsyncRndvCall
// @brief Post send a register MR with op info to peer, and waiting for response. Peer will be trigger new
//
//	request NetCallback also with op info. Poll mode:worker poll(default mode, support sync/async with sem), self
//	poll(support sync only, async is forbidden)
//
// @param channel       [in] net channel
// @param opInfo        [in] peer can get opInfo from ServiceGetOpInfo
// @param req           [in] send local MR to remote for data message, rKey/rAddress do not need to assign
// @param callback      [in] receive message will store in NetServiceContext::Message() in callback lifetime. If
// user want to use Message() in other thread, need to prepare buff and copy data out
//
// note: peer rsp message need invoke Send() with Service_OpInfo+
//
// @return 0 means the operation is in progress, cb will be invoked. return ERROR, cb will not be invoked
func ChannelAsyncRndvCall(channel NetChannel, opInfo ServiceOpInfo, req ServiceRequest, callback ChannelCallback) int {
	if callback.Cb == nil {
		fmt.Println("callback could not be null in async call")
		return -1
	}
	var cReq C.Service_Request
	cReq.lAddress = C.uintptr_t(req.LAddress)
	cReq.rAddress = C.uintptr_t(req.RAddress)
	cReq.lKey = C.uint32_t(req.LKey)
	cReq.rKey = C.uint32_t(req.RKey)
	cReq.size = C.uint32_t(req.Size)
	var cOpInfo C.Service_OpInfo
	cOpInfo.opCode = C.uint16_t(opInfo.OpCode)
	cOpInfo.timeout = C.int16_t(opInfo.Timeout)
	cOpInfo.flags = C.uint8_t(opInfo.Flags)
	cOpInfo.errorCode = C.int16_t(opInfo.ErrorCode)
	var cb C.Channel_Callback
	cb.cb = C.Channel_CallbackFunc(callback.Cb)
	cb.arg = callback.Arg
	res := C.Channel_AsyncRndvCall(C.Net_Channel(channel), &cOpInfo, &cReq, &cb)
	return int(res)
}

// PrepareRndvRequest
// @brief Prepare the rndv request for the communication
//
// @param allocator        [in] memory allocator
// @param msg              [in] the address of msg
// @param size             [in] size of the msg
//
// @return 0 means OK, other value means failed
// @return the prepared request
func PrepareRndvRequest(allocator MemoryAllocator, msg unsafe.Pointer, size uint64) (int, ServiceRequest) {
	var cAddress C.uintptr_t
	var cKey C.uint32_t
	res := C.Net_MemoryAllocatorAllocate(C.Net_MemoryAllocator(allocator), C.uint64_t(size), &cAddress, &cKey)

	var request ServiceRequest
	if res != 0 {
		return int(res), request
	}

    res = C.Cgo_MemoryCopy(unsafe.Pointer(uintptr(cAddress)), msg, C.uint32_t(size), C.uint32_t(size))
    if res != 0 {
        return int(res), request
    }

	request.LAddress = uintptr(cAddress)
	request.LKey = uint32(cKey)
	request.Size = uint32(size)
	return int(res), request
}

// ChannelRead
// @brief Post a single side read request to peer, no NetCallback at peer will be triggered
//
// @param channel        [in] net channel
// @param req            [in] request information, including 5 important variables, local/remote address/key and size.
// You can free it after called NetServiceOneSideDoneHandler.
// @param callback       [in] NetCallback for async send to peer. If cb is nullptr, the interface will block until
// complete. And it means if set cb is nullptr in worker callback context, the interface will block worker thread
//
// @return 0 if cb is null, it means the operation is cb successfully. Otherwise, it means the operation is in progress
func ChannelRead(channel NetChannel, req ServiceRequest, callback ChannelCallback) int {
	var res C.int
	var cRequest C.Service_Request
	cRequest.lAddress = C.uintptr_t(req.LAddress)
	cRequest.rAddress = C.uintptr_t(req.RAddress)
	cRequest.lKey = C.uint32_t(req.LKey)
	cRequest.rKey = C.uint32_t(req.RKey)
	cRequest.size = C.uint32_t(req.Size)
	if callback.Cb == nil {
		res = C.Channel_Read(C.Net_Channel(channel), &cRequest, nil)
	} else {
		var cb C.Channel_Callback
		cb.cb = C.Channel_CallbackFunc(callback.Cb)
		cb.arg = callback.Arg
		res = C.Channel_Read(C.Net_Channel(channel), &cRequest, &cb)
	}
	return int(res)
}

// ChannelWrite
// @brief Post a single side write request to peer, no NetCallback at peer will be triggered
//
// @param channel        [in] net channel
// @param req            [in] request information, including 5 important variables, local/remote address/key and size.
// You can free it after called NetServiceOneSideDoneHandler.
// @param callback       [in] NetCallback for async send to peer. If cb is nullptr, the interface will block until
// complete. And it means if set cb is nullptr in worker callback context, the interface will block worker thread
//
// @return 0 if cb is null, it means the operation is cb successfully. Otherwise, it means the operation is in progress
func ChannelWrite(channel NetChannel, req ServiceRequest, callback ChannelCallback) int {
	var res C.int
	var cRequest C.Service_Request
	cRequest.lAddress = C.uintptr_t(req.LAddress)
	cRequest.rAddress = C.uintptr_t(req.RAddress)
	cRequest.lKey = C.uint32_t(req.LKey)
	cRequest.rKey = C.uint32_t(req.RKey)
	cRequest.size = C.uint32_t(req.Size)
	if callback.Cb == nil {
		res = C.Channel_Write(C.Net_Channel(channel), &cRequest, nil)
	} else {
		var cb C.Channel_Callback
		cb.cb = C.Channel_CallbackFunc(callback.Cb)
		cb.arg = callback.Arg
		res = C.Channel_Write(C.Net_Channel(channel), &cRequest, &cb)
	}
	return int(res)
}

// ServiceRegisterMemoryRegion
// @brief Register a memory region, the memory will be allocated internally
//
// @param service          [in] the address of service
// @param size             [in] size of the memory region
//
// @return 0 means OK, other value means failed
// @return memory region registered
func ServiceRegisterMemoryRegion(service NetService, size uint64) (int, MemoryRegion) {
	var cMemoryRegion C.Service_MemoryRegion
	res := C.Service_RegisterMemoryRegion(C.Net_Service(service), C.uint64_t(size), &cMemoryRegion)
	return int(res), MemoryRegion(cMemoryRegion)
}

// ServiceRegisterAssignMemoryRegion
// @brief Register a memory region, the memory need to be passed in
//
// @param service        [in] the address of service
// @param address        [in] the memory point need to be registered
// @param size           [in] size of the memory region
//
// @return 0 means OK, other value means failed
// @return memory region registered
func ServiceRegisterAssignMemoryRegion(service NetService, memory Memory, size uint64) (int, MemoryRegion) {
	var cMemoryRegion C.Service_MemoryRegion
	res := C.Service_RegisterAssignMemoryRegion(C.Net_Service(service), C.uintptr_t(memory), C.uint64_t(size), &cMemoryRegion)
	return int(res), MemoryRegion(cMemoryRegion)
}

// ServiceGetMemoryRegionInfo
// @brief Parse the memory region, get info
//
// @param mr             [in] memory region registered
//
// @return 0 means OK, other value means failed
// @return memory region info
func ServiceGetMemoryRegionInfo(memoryRegion MemoryRegion) (int, ServiceMemoryRegionInfo) {
	var cMemoryRegionInfo C.Service_MemoryRegionInfo
	var memoryRegionInfo ServiceMemoryRegionInfo
	res := C.Service_GetMemoryRegionInfo(C.Service_MemoryRegion(memoryRegion), &cMemoryRegionInfo)
	if res != 0 {
		return int(res), memoryRegionInfo
	}
	memoryRegionInfo.LAddress = uintptr(cMemoryRegionInfo.lAddress)
	memoryRegionInfo.LKey = uint32(cMemoryRegionInfo.lKey)
	memoryRegionInfo.Size = uint32(cMemoryRegionInfo.size)
	return int(res), memoryRegionInfo
}

// ServiceDestroyMemoryRegion
// @brief Destroy the memory region
//
// @param service      			 [in] the address of service
// @param memoryRegion           [in] memory region registered
func ServiceDestroyMemoryRegion(service NetService, memoryRegion MemoryRegion) {
	C.Service_DestroyMemoryRegion(C.Net_Service(service), C.Net_Service(memoryRegion))
}

// EstimatedEncryptLen
// @brief Estimated Encrypt length for input raw len, only need to be use when one side communication by rdma is used
//
// @param channel          [in] the channel
// @param rawLen       	   [in] raw length before encrypt
//
// @return the length after encrypt
func EstimatedEncryptLen(channel NetChannel, rawLen uint64) uint64 {
	return uint64(C.Service_EstimatedEncryptLen(C.Net_Channel(channel), C.uint64_t(rawLen)))
}

// Encrypt
// @brief Encrypt data, only need to be use when one side communication by rdma is used
//
// @param channel          [in] the channel address
// @param rawData          [in] raw data before encrypt
// @param rawLen           [in] raw data length before encrypt
//
// @return 0 means OK, other value means failed
// @return cipher data after encrypt
// @return cipher data length after encrypt
func Encrypt(channel NetChannel, rawData MessageData, rawLen uint64) (int, MessageData, uint64) {
	var cipher unsafe.Pointer
	var cipherLen C.uint64_t
	res := C.Service_Encrypt(C.Net_Channel(channel), unsafe.Pointer(rawData), C.uint64_t(rawLen), cipher, &cipherLen)
	return int(res), MessageData(cipher), uint64(cipherLen)
}

// GetEncryptMsg
// @brief get Encrypt msg, only need to be use when one side communication by rdma is used
//
// @param channel          [in] the channel address
// @param rawData          [in] raw data before encrypt
// @param mrAddress        [in] mr address
// @param mrSize           [in] mr size
//
// @return 0 means OK, other value means failed
// @return cipher data length after encrypt
func GetEncryptMsg(channel NetChannel, rawData MessageData, rawLen uint64, mrAddress Memory, mrSize uint64) (int, uint64) {
	cipherLen := C.Cgo_GetEncryptMsg(C.Net_Channel(channel), unsafe.Pointer(rawData), C.uint64_t(rawLen), C.uintptr_t(mrAddress), C.uint64_t(mrSize))
	if cipherLen == 0 {
		return -1, 0
	}
	return 0, uint64(cipherLen)
}

// EstimatedDecryptLen
// @brief Estimate Decrypt length, only need to be use when one side communication by rdma is used
//
// @param channel          [in] the channel address
// @param cipherLen        [in] cipher len before decrypt
//
// @return the raw length after decrypt
func EstimatedDecryptLen(channel NetChannel, cipherLen uint64) uint64 {
	return uint64(C.Service_EstimatedDecryptLen(C.Net_Channel(channel), C.uint64_t(cipherLen)))
}

// Decrypt
// @brief Decrypt data, only need to be use when one side communication by rdma is used
//
// @param channel          [in] the channel address
// @param cipher           [in] cipher data after encrypt
// @param cipherLen        [in] cipher data length after encrypt
// @param rawData          [out] raw data
// @param rawLen           [in] raw data length
//
// @return 0 means OK, other value means failed
// @return actual raw data length
func Decrypt(channel NetChannel, cipher Memory, cipherLen uint64, rawData MemoryPointer, rawLen uint64) (int, uint64) {
	actualRawLen := C.Cgo_GetDecryptMsg(C.Net_Channel(channel), C.uintptr_t(cipher), C.uint64_t(cipherLen), unsafe.Pointer(rawData), C.uint64_t(rawLen))
	if actualRawLen == 0 {
		return -1, 0
	}
	return 0, uint64(actualRawLen)
}