/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2021-2021. All rights reserved.
 * Description:
 * Author: bao,
 * Create:
 */
#ifndef OCK_HCOM_CPP_UTIL_H_34523
#define OCK_HCOM_CPP_UTIL_H_34523

#include <cstdint>
#include <iostream>

#include "hcom_num_def.h"

namespace ock {
namespace hcom {
#define NET_FLAGS_BIT(i) (1UL << (i))

#define NET_SGE_MAX_IOV 4

constexpr const uint32_t NET_SGE_MAX_SIZE = 524288000;
constexpr const uint32_t NET_STR_ERROR_BUF_SIZE = 128;

// enum num should less than 128
enum NET_FLAGS {
    NTH_TWO_SIDE = 0,
    NTH_TWO_SIDE_SGL = 1,
    NTH_REPLY_REQUIRED = 2,
    NTH_READ = 3,
    NTH_READ_ACK = 4,
    NTH_WRITE = 5,
    NTH_WRITE_ACK = 6,
    NTH_READ_SGL = 7,
    NTH_READ_SGL_ACK = 8,
    NTH_WRITE_SGL = 9,
    NTH_WRITE_SGL_ACK = 10
};

#if __GNUC__ == 4 && __GNUC_MINOR__ == 8 && __GNUC_PATCHLEVEL__ == 5
template <class T, class U = T> T exchangeHcom(T &obj, U &&new_value)
{
    T old_value = std::move(obj);
    obj = std::forward<U>(new_value);
    return old_value;
}
#endif

template <typename T> class NetAtomicState {
public:
    explicit NetAtomicState(T state) : mState(state) {}

    ~NetAtomicState() = default;

    inline void Set(T newState)
    {
        __sync_lock_test_and_set(&mState, newState);
    }

    inline bool CAS(T oldState, T newState)
    {
        return __sync_bool_compare_and_swap(&mState, oldState, newState);
    }

    inline bool Compare(T state) const
    {
        return mState == state;
    }

    inline T Get() const
    {
        return mState;
    }

    NetAtomicState() = default;

    NetAtomicState(const NetAtomicState<T> &) = delete;

    NetAtomicState(NetAtomicState<T> &&) = delete;

    NetAtomicState<T> &operator = (const NetAtomicState<T> &) = delete;

    NetAtomicState<T> &operator = (NetAtomicState<T> &&) = delete;

private:
    T mState;
};

/**
 * @brief const variables
 */
const std::string CONST_EMPTY_STRING;

using NResult = int32_t;

constexpr uint16_t INVALID_WORKER_INDEX = 0xFFFF;
constexpr uint8_t INVALID_WORKER_GROUP_INDEX = 0xFF;
constexpr uint32_t INVALID_IP = 0xFFFFFFFF;

struct NetTransSgeIov {
    uintptr_t lAddress = 0;
    uintptr_t rAddress = 0;
    uint32_t lKey = 0;
    uint32_t rKey = 0;
    uint32_t size = 0;

    NetTransSgeIov() = default;
    NetTransSgeIov(uintptr_t lAddress, uintptr_t rAddress, uint32_t lKey, uint32_t rKey, uint32_t size)
        : lAddress(lAddress), rAddress(rAddress), lKey(lKey), rKey(rKey), size(size)
    {}
    NetTransSgeIov(uintptr_t lAddr, uint32_t lK, uint32_t s) : lAddress(lAddr), lKey(lK), size(s) {}
} __attribute__((packed));

struct NetTransHeader {
    uint32_t headerCrc = 0;  // header crc code
    int16_t opCode = 0;      // user define op code, it can be 0~1023
    uint16_t flags = 0;      // flags on the header, the upper 8 bits are reserved for the user
    uint32_t seqNo = 0;      // seq no
    int16_t timeout = 0;    // timeout from client
    int16_t errorCode = 0;   // error code for response
    uint32_t dataLength = 0; // body length
    uint32_t immData = 0;    // immData
    uintptr_t rAddress = 0;  // simulate rdma operation

    NetTransHeader() = default;

    inline void Invalid()
    {
        opCode = NN_NO1024;
        seqNo = NN_NO0;
        errorCode = NN_NO0;
        dataLength = NN_NO0;
    }
} __attribute__((packed));

/**
 * @brief Device information for user
 */
struct NetDriverDeviceInfo {
    int maxSge = NN_NO4; // max iov count in NetTransSglRequest
} __attribute__((packed));

/**
 * @brief macros
 */
//  macro for reference count
#ifndef DEFINE_RDMA_REF_COUNT_FUNCTIONS
#define DEFINE_RDMA_REF_COUNT_FUNCTIONS                    \
public:                                                    \
    inline void IncreaseRef()                              \
    {                                                      \
        __sync_fetch_and_add(&mRefCount, 1);               \
    }                                                      \
                                                           \
    inline void DecreaseRef()                              \
    {                                                      \
        int32_t tmp = __sync_sub_and_fetch(&mRefCount, 1); \
        if (tmp == 0) {                                    \
            delete this;                                   \
        }                                                  \
    }                                                      \
                                                           \
    inline int32_t GetRef()                                \
    {                                                      \
        return __sync_sub_and_fetch(&mRefCount, 0);        \
    }
#endif

#ifndef DEFINE_RDMA_REF_COUNT_VARIABLE
#define DEFINE_RDMA_REF_COUNT_VARIABLE \
private:                               \
    int32_t mRefCount = 0;
#endif

// macro for gcc optimization for prediction of if/else
#ifndef NN_LIKELY
#define NN_LIKELY(x) (__builtin_expect(!!(x), 1) != 0)
#endif

#ifndef NN_UNLIKELY
#define NN_UNLIKELY(x) (__builtin_expect(!!(x), 0) != 0)
#endif

// macro for flag
#define NN_FLAGS_BIT(i) (1UL << (i))
}
}

#endif // OCK_HCOM_CPP_UTIL_H_34523
