/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2021-2021. All rights reserved.
 * Author: bao
 */
#ifndef OCK_COMM_LOG_12456845341233_H
#define OCK_COMM_LOG_12456845341233_H

#include <ctime>
#include <mutex>
#include <iostream>
#include <sstream>
#include <sys/time.h>
#include <cstring>
#include "hcom_def.h"

namespace ock {
namespace hcom {
typedef void (*ExternalLog)(int level, const char *msg);

class NetOutLogger {
public:
    static NetOutLogger *Instance()
    {
        if (NN_UNLIKELY(gLogger == nullptr)) {
            std::lock_guard<std::mutex> lock(gMutex);
            if (gLogger == nullptr) {
                gLogger = new (std::nothrow) NetOutLogger();
                if (gLogger == nullptr) {
                    std::cout << "Failed to new NetOutLogger, probably out of memory" << std::endl;
                }
                SetLogLevel();
            }
        }

        return gLogger;
    }

    static void SetLogLevel()
    {
        /* set one of 0,1,2,3 */
        char *envSize = ::getenv("HCOM_SET_LOG_LEVEL");
        if (envSize != nullptr) {
            long value = 0;
            if (!SetStrStol(envSize, value)) {
                std::cout << "Invalid setting 'HCOM_SET_LOG_LEVEL', should set one of 0,1,2,3 " << std::endl;
                return;
            }
            logLevel = value;
        }
    }

    inline static void SetLogLevel(int level)
    {
        if (level >= static_cast<int>(NN_NO0) && level <= static_cast<int>(NN_NO3)) {
            logLevel = level;
        }
    }

    static bool SetStrStol(const std::string &str, long &value)
    {
        char *remain = nullptr;
        errno = 0;
        value = std::strtol(str.c_str(), &remain, 10); // 10 is decimal digits
        if (remain == nullptr || strlen(remain) > 0 || value < NN_NO0 || value > NN_NO3 || errno == ERANGE) {
            return false;
        } else if (value == 0 && str != "0") {
            return false;
        }

        return true;
    }

    inline void SetExternalLogFunction(ExternalLog func)
    {
        mLogFunc = func;
    }

    static inline void Print(int level, const char *msg)
    {
        struct timeval tv {};
        char strTime[24];

        int ret = gettimeofday(&tv, nullptr);
        if (ret != 0) {
            std::cout << "Fail to get the current system time, " << ret << "." << std::endl;
        }
        time_t timeStamp = tv.tv_sec;
        struct tm localTime {};
        struct tm *resultTime = localtime_r(&timeStamp, &localTime);
        if ((resultTime != nullptr) &&
            (strftime(strTime, sizeof strTime, "%Y-%m-%d %H:%M:%S.", resultTime) != NN_NO0)) {
            std::cout << strTime << tv.tv_usec << " " << level << " " << msg << std::endl;
        } else {
            std::cout << "Invalid time trace " << tv.tv_usec << " " << level << " " << msg << std::endl;
        }
    }

    inline void Log(int level, const std::ostringstream &oss) const
    {
        if (NN_LIKELY(mLogFunc != nullptr)) {
            mLogFunc(level, oss.str().c_str());
        } else {
            Print(level, oss.str().c_str());
        }
    }

    NetOutLogger(const NetOutLogger &) = delete;
    NetOutLogger &operator = (const NetOutLogger &) = delete;
    NetOutLogger(NetOutLogger &&) = delete;
    NetOutLogger &operator = (NetOutLogger &&) = delete;

    ~NetOutLogger()
    {
        mLogFunc = nullptr;
    }

    inline int GetLogLevel()
    {
        return logLevel;
    }

private:
    NetOutLogger() = default;

private:
    static NetOutLogger *gLogger;
    static std::mutex gMutex;
    static int logLevel;

    ExternalLog mLogFunc = nullptr;
};

// macro for log
#ifndef NN_LOG_FILENAME
#define NN_LOG_FILENAME (strrchr(__FILE__, '/') ? strrchr(__FILE__, '/') + 1 : __FILE__)
#endif

#define NN_LOG(level, args)                                                        \
    do {                                                                           \
        if ((level) >= (NetOutLogger::Instance()->GetLogLevel())) {                \
            std::ostringstream oss;                                                \
            oss << "[HCOM " << NN_LOG_FILENAME << ":" << __LINE__ << "] " << args; \
            NetOutLogger::Instance()->Log(level, oss);                             \
        }                                                                          \
    } while (0)

#define NN_LOG_PRINT(level, args)                                                  \
    do {                                                                           \
        if ((level) >= (NetOutLogger::Instance()->GetLogLevel())) {                \
            std::ostringstream oss;                                                \
            oss << "[HCOM " << NN_LOG_FILENAME << ":" << __LINE__ << "] " << args; \
            NetOutLogger::Instance()->Print(level, oss.str().c_str());             \
        }                                                                          \
    } while (0)

#define NN_LOG_DEBUG(args) NN_LOG(0, args)
#define NN_LOG_INFO(args) NN_LOG(1, args)
#define NN_LOG_WARN(args) NN_LOG(2, args)
#define NN_LOG_ERROR(args) NN_LOG(3, args)

#define NN_ASSERT_LOG_RETURN(args, RET)   \
    if (NN_UNLIKELY(!(args))) {           \
        NN_LOG_ERROR("Assert " << #args); \
        return RET;                       \
    }

#define NN_ASSERT_LOG_RETURN_VOID(args)   \
    if (NN_UNLIKELY(!(args))) {           \
        NN_LOG_ERROR("Assert " << #args); \
        return;                           \
    }

#ifdef NN_LOG_TRACE_INFO_ENABLED
#define NN_LOG_TRACE_INFO(args) NN_LOG_INFO(args)
#else
#define NN_LOG_TRACE_INFO(x)
#endif
}
}

#endif // OCK_COMM_LOG_12456845341233_H
