/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2021-2023. All rights reserved.
 * Author: bao, chen, fu, liu, ye
 */
#ifndef OCK_HCOM_SERVICE_H
#define OCK_HCOM_SERVICE_H

#include "hcom.h"

namespace ock {
namespace hcom {
class NetServiceContext;
class NetChannel;
class MultiRailNetChannel;
class NetService;
class NetServiceDefaultImp;
class NetCallback;
class NetServiceGlobalObject;
class NetServiceTimer;
class NetPeriodicManager;

using SerResult = int;

using NetChannelPtr = NetRef<NetChannel>;
using MultiRailNetChannelPtr = NetRef<MultiRailNetChannel>;
using NetDriverPtr = NetRef<NetDriver>;

enum NetChannelBrokenPolicy : uint8_t {
    BROKEN_ALL = 0, /* when one ep broken, all eps broken */
    RECONNECT,      /* when one ep broken, try re-connect first. If re-connect fail, broken all eps */
    KEEP_ALIVE,     /* when one ep broken, keep left eps alive until all eps broken */
};

using NetServiceNewChannelHandler =
    std::function<int(const std::string &ipPort, const NetChannelPtr &, const std::string &payload)>;
using NetServiceChannelBrokenHandler = std::function<void(const NetChannelPtr &)>;

using NetServiceReceivedHandler = std::function<int(NetServiceContext &)>;
using NetServiceSentHandler = std::function<int(const NetServiceContext &)>;
using NetServiceOneSideDoneHandler = std::function<int(const NetServiceContext &)>;
using NetServiceIdleHandler = NetDriverIdleHandler;

/**
 * @brief Message will be sent to peer or received from peer
 */
struct NetServiceMessage {
    void *data = nullptr;       /* pointer of data */
    uint32_t size = 0;          /* size of data */
    bool transferOwner = false; /* reserved, transfer data ownership to hcom, it will be freed after transferred */

    NetServiceMessage() = default;

    /**
     * @brief Constructor
     * @param d                [in] pointer of data to be sent or received
     * @param s                [in] size of data to be sent or received
     */
    NetServiceMessage(void *d, uint32_t s) : data(d), size(s) {}
} __attribute__((packed));

/**
 * @brief Request info of one side operation
 */
struct NetServiceRequest {
    uintptr_t lAddress = 0; /* data address of local */
    uintptr_t rAddress = 0; /* data address of remote */
    uint32_t lKey = 0;      /* key of local memory region */
    uint32_t rKey = 0;      /* key of remote memory region */
    uint32_t size = 0;      /* size of data to be operated */

    NetServiceRequest() = default;
    NetServiceRequest(uintptr_t la, uintptr_t ra, uint32_t lk, uint32_t rk, uint32_t s)
        : lAddress(la), rAddress(ra), lKey(lk), rKey(rk), size(s)
    {}
} __attribute__((packed));

/**
 * @brief Sgl operation request
 */
struct NetServiceSglRequest {
    NetServiceRequest *iov = nullptr; /* based address of sge array */
    uint16_t iovCount = 0;            /* count of Sge vector, max count: NET_SGE_MAX_IOV */

    NetServiceSglRequest() = default;
    NetServiceSglRequest(NetServiceRequest *iovPtr, uint16_t cnt) : iov(iovPtr), iovCount(cnt) {}

    inline uint64_t Size() const
    {
        if (NN_UNLIKELY(iovCount > NET_SGE_MAX_IOV || iov == nullptr)) {
            NN_LOG_ERROR("Invalid iov count " << iov << " or iov ptr " << iov);
            return NN_NO0;
        }

        uint64_t size = 0;
        for (uint16_t i = 0; i < iovCount; i++) {
            size += iov[i].size;
        }
        return size;
    }
} __attribute__((packed));

/**
 * @brief Two side message operation info, include operation code, timeout, reply error code etc
 */
struct NetServiceOpInfo {
    uint16_t opCode = 0;   /* operation code, 0~999 */
    int16_t timeout = -1;  /* timeout in seconds, set current operation */
    int16_t errorCode = 0; /* error code, usually set by response side */
    uint8_t flags = 0;     /* flags for user */

    NetServiceOpInfo() = default;
    explicit NetServiceOpInfo(uint16_t op, int16_t timeout = -1, int16_t errorCode = 0)
        : opCode(op), timeout(timeout), errorCode(errorCode)
    {}
    NetServiceOpInfo(uint16_t op, uint8_t flags, int16_t timeout, int16_t errorCode)
        : opCode(op), timeout(timeout), errorCode(errorCode), flags(flags)
    {}

    inline void Invalidate()
    {
        opCode = NN_NO1024;
        timeout = NN_NO0;
        flags = NN_NO0;
        errorCode = NN_NO0;
    }
} __attribute__((packed));

struct NetServiceMrKey {
    uint8_t idx = 0;   /* idx in NetDriver::mIndex */
    uint32_t lKey = 0; /* key of local memory region */
} __attribute__((packed));

struct NetServiceMemoryRegionKey {
    uint32_t count = 0;            /* count of register MemoryRegion key array */
    NetServiceMrKey mrKey[NN_NO4]; /* register MemoryRegion key array */

    int32_t FindMrKey(uint8_t driverIndex)
    {
        for (uint32_t i = 0; i < count; ++i) {
            if (mrKey[i].idx == driverIndex) {
                return mrKey[i].lKey;
            }
        }
        return -1;
    }
} __attribute__((packed));

/**
 * Type index for channel, used for connecting and NetCallback identification
 * for example, ctrl channel and data channel could be 0 and 1
 */
using NetChTypeIndex = uint16_t;

/**
 * @brief Seq number
 */
union NetSeqNo {
    struct {
        /* low address */
        uint32_t realSeq : 24; /* real seq no */
        uint32_t version : 6;  /* request version */
        uint32_t fromFlat : 1; /* allocated from flat or hash map */
        uint32_t isResp : 1;   /* request or reply, 0 for request, 1 for reply */
        /* high address */
    };
    uint32_t wholeSeq = 0;

    explicit NetSeqNo(uint32_t whole) : wholeSeq(whole) {}

    inline void SetValue(uint32_t flat, uint32_t ver, uint32_t seq)
    {
        fromFlat = flat;
        version = ver;
        realSeq = seq;
    }

    std::string ToString() const
    {
        std::ostringstream oss;
        oss << "NetSeqNo info=[wholeSeq: " << wholeSeq << ", isResp: " << isResp << ", fromFlat: " << fromFlat <<
            ", version: " << version << ", realSeq: " << realSeq << "]";
        return oss.str();
    }

    inline bool IsResp() const
    {
        return isResp == NN_NO1;
    }
};

/**
 * Endpoint to channel upCtx, store ch pointer and endpoint index into uint64_t
 */
union Ep2ChUpCtx {
    struct {
        uint64_t connected : 1; /* flag for connecting or connected, store different type ptr */
        uint64_t epIdx : 5;     /* endpoint index, range [0, 31] */
        uint64_t ptr : 58;      /* pointer to connecting mgr or net channel */
    };
    uint64_t wholeUpCtx = 0; /* whole */

    Ep2ChUpCtx() = default;
    explicit Ep2ChUpCtx(uint64_t w) : wholeUpCtx(w) {}
    Ep2ChUpCtx(uint64_t connect, uint64_t p, uint64_t i) : connected(connect), epIdx(i), ptr(p) {}

    inline NetChannel *Channel() const
    {
        if (NN_UNLIKELY(connected != NN_NO1)) {
            NN_LOG_ERROR("Failed to get channel by not connected");
            return nullptr;
        }
        return reinterpret_cast<NetChannel *>(ptr);
    }

    inline uint64_t Ptr() const
    {
        return ptr;
    }

    inline uint32_t EpIdx() const
    {
        return static_cast<uint32_t>(epIdx);
    }

    std::string ToString() const
    {
        std::ostringstream oss;
        oss << "chPtr " << ptr << ", epIdx " << epIdx << ", whole " << wholeUpCtx;
        return oss.str();
    }
};

/**
 * MultiRail channel upCtx, store MultiRail channel pointer and channel index into uint64_t
 */
union MultiRailChUpCtx {
    struct {
        uint64_t connected : 1;  /* flag for connecting or connected, store different type ptr */
        uint64_t channelIdx : 5; /* endpoint index, range [0, 31] */
        uint64_t ptr : 58;       /* pointer to connecting mgr or net channel */
    };
    uint64_t wholeUpCtx = 0; /* whole */

    MultiRailChUpCtx() = default;
    explicit MultiRailChUpCtx(uint64_t w) : wholeUpCtx(w) {}
    MultiRailChUpCtx(uint64_t connect, uint64_t p, uint64_t i) : connected(connect), channelIdx(i), ptr(p) {}

    inline MultiRailNetChannel *Channel() const
    {
        if (NN_UNLIKELY(connected != NN_NO1)) {
            NN_LOG_ERROR("Failed to get channel by not connected");
            return nullptr;
        }
        return reinterpret_cast<MultiRailNetChannel *>(ptr);
    }

    inline uint64_t Ptr() const
    {
        return ptr;
    }

    inline uint32_t ChannelIdx() const
    {
        return static_cast<uint32_t>(channelIdx);
    }

    std::string ToString() const
    {
        std::ostringstream oss;
        oss << "chPtr " << ptr << ", channelIdx " << channelIdx << ", whole " << wholeUpCtx;
        return oss.str();
    }
};

/* *
 * @brief Context of request received, operationInfo/message/channel can be got from it,
 * and reply message with it
 */
class NetServiceContext {
public:
    enum Operation : uint8_t {
        SER_RECEIVED = 0,     /* support invoke all functions */
        SER_RECEIVED_RAW = 1, /* support invoke most functions except OpInfo() */
        SER_SENT = 2,         /* support invoke basic functions except
                                 Message()、MessageData()、MessageDataLen()、RspCtx()、all ReplySend*() */
        SER_SENT_RAW = 3,     /* support invoke basic functions except
                                 Message()、MessageData()、MessageDataLen()、RspCtx()、OpInfo()、all ReplySend*() */
        SER_ONE_SIDE = 4,     /* support invoke basic functions except
                                 Message()、MessageData()、MessageDataLen()、RspCtx()、OpInfo()、all ReplySend*() */

        SER_INVALID_OP_TYPE = 255,
    };
    /**
     * @brief Get result of all operation
     */
    SerResult Result() const;

    /**
     * @brief Get the channel ptr
     */
    const NetChannelPtr &Channel() const;

    /**
     * @brief Get the operation type
     * @return SER_INVALID_OP_TYPE if failed
     */
    Operation OpType() const;

    /**
     * @brief Get response context for send rsp message in other thread
     * note: only support SER_RECEIVED/SER_RECEIVED_RAW invoke
     * @return ture if success, false if failed
     */
    uintptr_t RspCtx() const;

    /**
     * @brief Reply send message to peer
     * note: only support SER_RECEIVED/SER_RECEIVED_RAW invoke
     * @return SER_OK if success, others if failed
     */
    SerResult ReplySend(const NetServiceOpInfo &opInfo, const NetServiceMessage &msg, const NetCallback *done);
    SerResult ReplySendRaw(const NetServiceMessage &msg, const NetCallback *done);
    SerResult ReplySendRawSgl(const NetServiceSglRequest &msg, const NetCallback *done);

    /**
     * @brief Get op info by user input
     * note: only support SER_SENT/SER_RECEIVED invoke
     */
    const NetServiceOpInfo &OpInfo();

    /**
     * @brief Get op code by user input
     * note: only support SER_SENT/SER_RECEIVED invoke
     * @return 0~999 if success, others if failed
     */
    uint16_t OpCode() const;

    /**
     * @brief Get the message data received which valid in callback lifetime
     * note1: only support SER_RECEIVED/SER_RECEIVED_RAW invoke
     * note2: if user want to use message in other thread, need to copy message.data by self or invoke clone()
     * @return valid address if success, nullptr if failed
     */
    void *MessageData() const;

    /**
     * @brief Get the message data received which valid in callback lifetime
     * note1: only support SER_RECEIVED/SER_RECEIVED_RAW invoke
     * @return valid length if success, 0 if failed
     */
    uint32_t MessageDataLen() const;

    /**
     * @brief clone service context
     * @param copyData : true means malloc and copy receive data.
     * @return SER_OK if success, others if failed
     */
    static SerResult Clone(NetServiceContext &newOne, const NetServiceContext &oldOne, bool copyData = true);

    /**
     * @brief check current context timeout or not
     * @return true if timeout, false if not timeout
     */
    bool IsTimeout() const;

    void Invalidate();

    ~NetServiceContext()
    {
        Invalidate();
    }

    NetServiceContext() = default;

private:
    NetServiceContext(const NetRequestContext &ctx, NetChannel *ch);

    enum DataType : uint8_t {
        OUTER_DATA = 0,    /* assign by NetRequestContext.Message()->Data() */
        MEM_POOL_DATA = 1, /* alloc from channel mem pool */

        INVALID_DATA = 255,
    };

    SerResult CopyData(void *data, uint32_t dataLen);

    /* 64B cache line align start */
    NetChannelPtr mCh;       /* channel ptr */
    int32_t mResult = 0;     /* context result */
    uint32_t mEpIdxInCh = 0; /* for response */
    uint32_t mSeqNo = 0;     /* for response */

    NetRequestContext::NN_OpType mOpType = NetRequestContext::NN_INVALID_OP_TYPE; /* operate original type */
    NetServiceOpInfo mOpInfo {};                                                  /* user sent/received op info */

    uint8_t rsv1[NN_NO3] {}; /* DataType occupies 1 byte */

    DataType mDateType = INVALID_DATA; /* type of mData */
    uint32_t mDataLen = 0;             /* received/received raw message data len */
    void *mData = nullptr;             /* received/received raw message data address */

    uint64_t mTimeoutTraceMs = 0; /* record timeout time trace, 0 means never timeout */
    uint32_t mReadCount = 0;
    uint8_t rsv2[NN_NO8] {};
    /* 64B cache line align end */

    friend class NetService;
    friend class NetChannel;
    friend class NetServiceDefaultImp;
    friend class NetServiceMultiRailImp;
    friend class NetServiceGlobalObject;
    friend class NetPeriodicManager;
};

inline SerResult NetServiceContext::Result() const
{
    return mResult;
}

inline const NetChannelPtr &NetServiceContext::Channel() const
{
    return mCh;
}

inline NetServiceContext::Operation NetServiceContext::OpType() const
{
    switch (mOpType) {
        case NetRequestContext::NN_SENT:
            return SER_SENT;
        case NetRequestContext::NN_SENT_RAW:
        case NetRequestContext::NN_SENT_RAW_SGL:
            return SER_SENT_RAW;
        case NetRequestContext::NN_RECEIVED:
            return SER_RECEIVED;
        case NetRequestContext::NN_RECEIVED_RAW:
            return SER_RECEIVED_RAW;
        case NetRequestContext::NN_WRITTEN:
        case NetRequestContext::NN_SGL_WRITTEN:
        case NetRequestContext::NN_READ:
        case NetRequestContext::NN_SGL_READ:
            return SER_ONE_SIDE;
        default:
            return SER_INVALID_OP_TYPE;
    }
}

inline uintptr_t NetServiceContext::RspCtx() const
{
    /* 8byte: low 4 byte for seq no, high 4 byte for epIndex */
    uintptr_t rsp = mEpIdxInCh;
    rsp = (rsp << NN_NO32) | mSeqNo;
    return rsp;
}

inline const NetServiceOpInfo &NetServiceContext::OpInfo()
{
    return mOpInfo;
}

inline uint16_t NetServiceContext::OpCode() const
{
    return mOpInfo.opCode;
}

inline void *NetServiceContext::MessageData() const
{
    return mData;
}

inline uint32_t NetServiceContext::MessageDataLen() const
{
    return mDataLen;
}

inline void NetServiceContext::Invalidate()
{
    if (mDateType == MEM_POOL_DATA && mData != nullptr) {
        free(mData);
        mData = nullptr;
        mDateType = INVALID_DATA;
    }
    mCh.Set(nullptr);
}

using NetServiceRndvAllocateHander = std::function<int(uint64_t size, uintptr_t &outAddress, uint32_t &outKey)>;
using NetServiceRndvFreeHander = std::function<int(uintptr_t address)>;

struct RndvMessage {
    uint16_t cnt = 0;
    NetServiceMessage msg[NET_SGE_MAX_IOV] {};

    RndvMessage() = default;
};

class NetServiceRndvContext {
public:
    enum RndvOpType : uint8_t {
        SER_RNDV = 0,     /* support invoke all functions */
        SER_RNDV_SGL = 1, /* support invoke all functions */
        SER_MULTIRAIL_RNDV_RAW = 2, /* support invoke all functions */
        SER_RNDV_INVALID_TYPE = 255,
    };

    inline RndvOpType OpType() const
    {
        return mType;
    }

    /**
     * @brief Get op info by user input, if invoke by SER_RNDV_SGL, return invalid
     * note: only support SER_RNDV invoke
     */
    inline const NetServiceOpInfo &OpInfo()
    {
        return mContext.OpInfo();
    }

    /**
     * @brief Get op code by user input, if invoke by SER_RNDV_SGL, return invalid
     * note: only support SER_RNDV invoke
     * @return 0~999 if success, others if failed
     */
    inline uint16_t OpCode() const
    {
        return mContext.OpCode();
    }

    /**
     * @brief check current context timeout or not
     * @return true if timeout, false if not timeout
     */
    bool IsTimeout() const;

    /**
     * @brief Get message
     * note: there may be one or more cnt by rndv call or rndv sgl call
     * @return RndvMessage
     */
    inline const RndvMessage &Message()
    {
        return mMessage;
    }

    /**
     * @brief Free NetServiceRndvContext
     * @return SER_OK if success, others if failed
     */
    int FreeMessage();

    /**
     * @brief Reply send message to peer
     * note: only support SER_RECEIVED/SER_RECEIVED_RAW invoke
     * @return SER_OK if success, others if failed
     */
    inline SerResult ReplyRndv(const NetServiceOpInfo &opInfo, const NetServiceMessage &msg, const NetCallback *done)
    {
        return mContext.ReplySend(opInfo, msg, done);
    }

    /**
     * @brief Get the channel ptr
     */
    inline const NetChannelPtr &Channel() const
    {
        return mContext.Channel();
    }

    ~NetServiceRndvContext()
    {
        (void)FreeMessage();
    }

    NetServiceRndvContext() = default;

private:
    NetServiceContext mContext {};
    RndvMessage mMessage {};
    NetServiceRndvFreeHander mFreeMethod = nullptr;
    uint64_t mTimestamp = 0;
    RndvOpType mType = SER_RNDV_INVALID_TYPE;
    uint32_t rndvCount = 0;

    friend class NetServiceDefaultImp;
    friend class NetServiceMultiRailImp;
};

class NetCallback {
public:
    NetCallback() = default;
    virtual ~NetCallback() = default;

    virtual void Run(NetServiceContext &context) = 0;
};

/**
 * @brief Closure NetCallback.
 *
 * @param ClosureFunction
 */
template <typename ClosureFunction> class ClosureCallback : public NetCallback {
public:
    explicit ClosureCallback(ClosureFunction &&function, bool deleteSelf) : mFunction(function), mDeleteSelf(deleteSelf)
    {}

    ~ClosureCallback() override = default;

    void Run(NetServiceContext &context) override
    {
        bool isDeleteSelf = false;
        if (mDeleteSelf) {
            mDeleteSelf = false;
            isDeleteSelf = true;
        }
        mFunction(context);
        if (isDeleteSelf) {
            delete this;
        }
    }

private:
    ClosureFunction mFunction = nullptr;
    bool mDeleteSelf = true;
};

/**
 * @brief Generate a self-deleting NetCallback object.
 *
 * @param Args
 * @param args
 * @return NetCallback*
 * @note At present, asynchronous operation is not a hot spot. In order to simplify
 * coding, std::bind is used to implement closure. If the cost of std::bind
 * is found to be high, then optimize it.
 */
template <typename... Args> NetCallback *NewCallback(Args... args)
{
    auto closure = std::bind(args...);
    return new (std::nothrow) ClosureCallback<decltype(closure)>(std::move(closure), true);
}

/**
 * @brief Generate a permanent callback object.
 *
 * @param Args
 * @param args
 * @return NetCallback*
 * @note see @ref NewCallback.
 */
template <typename... Args> NetCallback *NewPermanentCallback(Args... args)
{
    auto closure = std::bind(args...);
    return new (std::nothrow) ClosureCallback<decltype(closure)>(std::move(closure), false);
}

struct TimerContext {
    uint32_t seqNo = 0;
    NetServiceTimer *timer = nullptr;

    TimerContext() = default;
};

/**
 * @brief channel callback type
 */
enum NetChannelCBType : uint8_t {
    CHANNEL_FUNC_CB = 0,   // use channel function param (const NetCallback *done)
    CHANNEL_GLOBAL_CB = 1, // use service RegisterOpHandler
};

struct NetServiceConnectOptions {
    NetChTypeIndex index = 0;                  /* connect type index for different op handlers */
    uint8_t epSize = 1;                        /* assign ep size in channel, valid range is [1, 16] */
    NetChannelCBType cbType = CHANNEL_FUNC_CB; /* channel callback mode */
    uint8_t clientGrpNo = 0;                   /* indicates which server worker group to connect to */
    uint8_t serverGrpNo = 0;                   /* indicates which client worker group to connect to */
    uint32_t flags = 0;                        /* for server is invalids, for client determines the sync ep type :
                                                  NET_EP_SELF_POLLING | NET_EP_EVENT_POLLING */
    NetServiceConnectOptions() = default;

    NetServiceConnectOptions(NetChTypeIndex idx, uint8_t ep, NetChannelCBType cb, uint8_t clientNO, uint8_t serverNO,
        uint32_t flag)
        : index(idx), epSize(ep), cbType(cb), clientGrpNo(clientNO), serverGrpNo(serverNO), flags(flag)
    {}

    bool Compare(const NetServiceConnectOptions &options) const;
} __attribute__((packed));

enum ChannelFlowCtrlLevel : uint8_t {
    HIGH_LEVEL_BLOCK, /* spin-wait by busy loop */
    LOW_LEVEL_BLOCK,  /* full sleep */
};

struct ChannelFlowCtrlOptions {
    ChannelFlowCtrlLevel level = LOW_LEVEL_BLOCK; /* block level by different action */
    uint16_t intervalTimeMs = 0;                  /* interval time ms, range in [1, 1000] */
    uint64_t thresholdByte = 0;                   /* threshold byte */
};

struct ConnectTimestamp {
    uint64_t localTimeUs = 0;  /* local time trace when connecting */
    uint64_t remoteTimeUs = 0; /* remote time trace when connecting */
    uint64_t deltaTimeUs = 0;  /* delta time for exchange info */

    ConnectTimestamp() = default;
    ConnectTimestamp(uint64_t lTime, uint64_t rTime, uint64_t dTime)
        : localTimeUs(lTime), remoteTimeUs(rTime), deltaTimeUs(dTime)
    {}

    uint64_t GetRemoteTimestamp(int16_t timeOutSecond) const;
};

#define ENABLE_MULTI_RAIL 1 // stub
#if ENABLE_MULTI_RAIL
#define VIRTUAL virtual
#else
#define VIRTUAL
#endif

class NetChannel {
public:
    /**
     * @brief Post send a request with op info to peer, and without waiting for response. Peer will be trigger
     * new request NetCallback also with op info. Poll mode:worker poll(default mode, support sync/async with sem),
     * self poll(support sync only)
     *
     * @param opInfo        [in] peer can get opInfo from NetServiceContext.OpInfo()
     * @param req           [in] rpc message
     * @param done          [in] NetCallback for async send to peer. If done is nullptr, the interface will block
     * until complete. And it means if set done is nullptr in worker callback context, the interface will block
     * worker thread.
     * @param rspCtx        [in] 0 means request message by default. Others value means response message, which can
     * get from NetServiceContext.RspCtx()
     *
     * @return 0 if done is nullptr, it means the operation is done successfully. Otherwise, it means the operation
     * is in progress.
     * return ERROR, done will not be invoked.
     *
     */
    VIRTUAL SerResult Send(const NetServiceOpInfo &opInfo, const NetServiceMessage &req, const NetCallback *done,
        uintptr_t rspCtx = 0);

    /**
     * @brief Post send a request without op inf to peer, and without waiting for response. Peer will be
     * trigger new request NetCallback also without op info. Poll mode:worker poll(default mode, support sync/async
     * with sem), self poll(support sync only)
     *
     * @param message       [in] rpc message.
     * @param done          [in] NetCallback for async send to peer. If done is nullptr, the interface will block
     * until complete. And it means if set done is nullptr in worker callback context, the interface will block
     * worker thread.
     * @param rspCtx        [in] 0 means request message by default. Others value means response message, which can
     * get from NetServiceContext.RspCtx()
     *
     * @return 0 if done is nullptr, it means the operation is done successfully. Otherwise, it means the operation
     * is in progress.
     * return ERROR, done will not be invoked.
     *
     */
    VIRTUAL SerResult SendRaw(const NetServiceMessage &req, const NetCallback *done, uintptr_t rspCtx = 0);

    /**
     * @brief Post send a request without op info to peer, and without waiting for response. Peer will be
     * trigger new request NetCallback also without op info, this could be used when you have self define
     * header. Poll mode:worker poll(default mode, support sync/async with sem), self
     * poll(support sync only)
     *
     * @param req          [in] request information, fill with local different MRs, send to the same remote MR by
     * local MRs sequence, you can free it after called NetServiceOneSideDoneHandler. rKey/rAddress do not need to
     * assign.
     * @param done         [in] NetCallback for async send to peer. If done is nullptr, the interface will block
     * until complete. And it means if set done is nullptr in worker callback context, the interface will block
     * worker thread.
     * @param rspCtx       [in] 0 means request message by default. Others value means response message, which can
     * get from NetServiceContext.RspCtx()
     *
     * @return 0 if done is nullptr, it means the operation is done successfully. Otherwise, it means the operation
     * is in progress.
     * return ERROR, done will not be invoked.
     *
     */
    VIRTUAL SerResult SendRawSgl(const NetServiceSglRequest &req, const NetCallback *done, uintptr_t rspCtx = 0);

    /**
     * @brief Post send a request with op info to peer, and waiting for response. Peer will be trigger new
     * request NetCallback also with op info. Poll mode:worker poll(default mode, support sync/async with sem), self
     * poll(support sync only)
     *
     * @param reqOpInfo     [in] peer can get opInfo from NetServiceContext.OpInfo()
     * @param req           [in] send to remote message.
     * @param rspOpInfo     [out] peer response op info
     * @param rsp           [out] receive remote response message. If response is not nullptr, it will receive a
     * response and set rsp->size to real size. If user does not know the real length of the response, user can set
     * rsp->data to nullptr, so RPC will allocate enough memory to hold the response, and store the memory address
     * to response->data. User are required to use free() to release this memory. There will be additional overhead.
     *
     * @return 0 means OK, other value means failed.
     *
     */
    VIRTUAL SerResult SyncCall(const NetServiceOpInfo &reqOpInfo, const NetServiceMessage &req,
        NetServiceOpInfo &rspOpInfo, NetServiceMessage &rsp);

    /**
     * @brief Post send a request with op info to peer, and waiting for response. Peer will be trigger new
     * request NetCallback also with op info. Poll mode:worker poll(default mode, support sync/async with sem), self
     * poll(support sync only, async is forbidden)
     *
     * @param opInfo        [in] peer can get opInfo from NetServiceContext.OpInfo()
     * @param req           [in] send to remote message.
     * @param done          [in] receive message will store in NetServiceContext::Message() in callback lifetime. If
     * user want to use Message() in other thread, need to prepare buff and copy data out.
     *
     * @return 0 means the operation is in progress, done will be invoked.
     * return ERROR, done will not be invoked.
     *
     */
    VIRTUAL SerResult AsyncCall(const NetServiceOpInfo &opInfo, const NetServiceMessage &req, const NetCallback *done);

    /**
     * @brief Post send a request without op info to peer, and waiting for response. Peer will be trigger new
     * request NetCallback also with op info. Poll mode:worker poll(default mode, support sync/async with sem), self
     * poll(support sync only)
     *
     * @param req           [in] send to remote message.
     * @param rsp           [in] receive remote response message. If response is not nullptr, it will receive a
     * response and set rsp->size to real size. If user does not know the real length of the response, user can set
     * rsp->data to nullptr, so RPC will allocate enough memory to hold the response, and store the memory address
     * to response->data. User are required to use free() to release this memory. There will be additional overhead.
     *
     * @return 0 means OK, other value means failed.
     *
     */
    VIRTUAL SerResult SyncCallRaw(const NetServiceMessage &req, NetServiceMessage &rsp);

    /**
     * @brief Post send a request without op info to peer, and waiting for response. Peer will be trigger new
     * request NetCallback also with op info. Poll mode:worker poll(default mode, support sync/async with sem), self
     * poll(support sync only, async is forbidden)
     *
     * @param req           [in] send to remote message.
     * @param done          [in] receive message will store in NetServiceContext::Message() in callback lifetime. If
     * user want to use Message() in other thread, need to prepare buff and copy data out.
     *
     * @return 0 means the operation is in progress, done will be invoked.
     * return ERROR, done will not be invoked.
     *
     */
    VIRTUAL SerResult AsyncCallRaw(const NetServiceMessage &req, const NetCallback *done);

    /**
     * @brief Post send a request without op info to peer, and waiting for response. Peer will be trigger new
     * request NetCallback also with op info. Poll mode:worker poll(default mode, support sync/async with sem), self
     * poll(support sync only)
     *
     * @param req           [in] request information, fill with local different MRs, send to the same remote MR by
     * local MRs sequence, you can free it after called. rKey/rAddress do not need to assign.
     * @param rsp           [in] receive remote response message. If response is not nullptr, it will receive a
     * response and set rsp->size to real size. If user does not know the real length of the response, user can set
     * rsp->data to nullptr, so RPC will allocate enough memory to hold the response, and store the memory address
     * to response->data. User are required to use free() to release this memory. There will be additional overhead.
     *
     * @return 0 means OK, other value means failed.
     *
     */
    VIRTUAL SerResult SyncCallRawSgl(const NetServiceSglRequest &req, NetServiceMessage &rsp);

    /**
     * @brief Post send a request without op info to peer, and waiting for response. Peer will be trigger new
     * request NetCallback also with op info. Poll mode:worker poll(default mode, support sync/async with sem), self
     * poll(support sync only, async is forbidden)
     *
     * @param req           [in] request information, fill with local different MRs, send to the same remote MR by
     * local MRs sequence, you can free it after called. rKey/rAddress do not need to assign.
     * @param done          [in] receive message will store in NetServiceContext::Message() in callback lifetime. If
     * user want to use Message() in other thread, need to prepare buff and copy data out.
     *
     * @return 0 means the operation is in progress, done will be invoked.
     * return ERROR, done will not be invoked.
     *
     */
    VIRTUAL SerResult AsyncCallRawSgl(const NetServiceSglRequest &req, const NetCallback *done);

    /**
     * @brief Post send a register MR with op info to peer, and waiting for response. Peer will be trigger new
     * request NetCallback also with op info. Poll mode:worker poll(default mode, support sync/async with sem), self
     * poll(support sync only)
     *
     * @param reqOpInfo     [in] peer can get opInfo from NetServiceContext.OpInfo()
     * @param req           [in] send local MR to remote for data message, rKey/rAddress do not need to assign
     * @param rspOpInfo     [out] peer response op info
     * @param rsp           [out] receive remote response message for control message, always used for check result. If
     * response is not nullptr, it will receive a response and set rsp->size to real size. If user does not know the
     * real length of the response, user can set rsp->data to nullptr, so RPC will allocate enough memory to hold the
     * response, and store the memory address to response->data. User are required to use free() to release this memory.
     * There will be additional overhead.
     *
     * note: peer rsp message need invoke Send() with NetServiceOpInfo
     *
     * @return 0 means OK, other value means failed.
     *
     */
    VIRTUAL SerResult SyncRndvCall(const NetServiceOpInfo &reqOpInfo, const NetServiceRequest &req,
        NetServiceOpInfo &rspOpInfo, NetServiceMessage &rsp);

    /**
     * @brief Post send a register MR with op info to peer, and waiting for response. Peer will be trigger new
     * request NetCallback also with op info. Poll mode:worker poll(default mode, support sync/async with sem), self
     * poll(support sync only, async is forbidden)
     *
     * @param opInfo        [in] peer can get opInfo from NetServiceContext.OpInfo()
     * @param req           [in] send local MR to remote for data message, rKey/rAddress do not need to assign
     * @param done          [in] receive message will store in NetServiceContext::Message() in callback lifetime. If
     * user want to use Message() in other thread, need to prepare buff and copy data out.
     *
     * note: peer rsp message need invoke Send() with NetServiceOpInfo
     *
     * @return 0 means the operation is in progress, done will be invoked.
     * return ERROR, done will not be invoked.
     *
     */
    VIRTUAL SerResult AsyncRndvCall(const NetServiceOpInfo &reqOpInfo, const NetServiceRequest &req,
        const NetCallback *done);

    /**
     * @brief Post send a register MRs to peer, and waiting for response. Peer will be trigger new
     * request NetCallback also with op info. Poll mode:worker poll(default mode, support sync/async with sem), self
     * poll(support sync only)
     *
     * @param reqOpInfo     [in] peer can get opInfo from NetServiceContext.OpInfo()
     * @param request       [in] send local MRs to remote for data message, rKey/rAddress do not need to assign
     * @param rspOpInfo     [out] peer response op info
     * @param rsp           [out] receive remote response message for control message, always used for check result. If
     * response is not nullptr, it will receive a response and set rsp->size to real size. If user does not know the
     * real length of the response, user can set rsp->data to nullptr, so RPC will allocate enough memory to hold the
     * response, and store the memory address to response->data. User are required to use free() to release this memory.
     * There will be additional overhead.
     *
     * note: peer rsp message need invoke Send() with NetServiceOpInfo
     *
     * @return 0 means OK, other value means failed.
     *
     */
    VIRTUAL SerResult SyncRndvSglCall(const NetServiceOpInfo &reqOpInfo, const NetServiceSglRequest &req,
        NetServiceOpInfo &rspOpInfo, NetServiceMessage &rsp);

    /**
     * @brief Post send a request without op info to peer, and waiting for response. Peer will be trigger new
     * request NetCallback also with op info. Poll mode:worker poll(default mode, support sync/async with sem), self
     * poll(support sync only, async is forbidden)
     *
     * @param req           [in] send local MRs to remote for data message, rKey/rAddress do not need to assign
     * @param done          [in] receive message will store in NetServiceContext::Message() in callback lifetime. If
     * user want to use Message() in other thread, need to prepare buff and copy data out.
     *
     * note: peer rsp message need invoke Send() with NetServiceOpInfo
     *
     * @return 0 means the operation is in progress, done will be invoked.
     * return ERROR, done will not be invoked.
     *
     */
    VIRTUAL SerResult AsyncRndvSglCall(const NetServiceOpInfo &reqOpInfo, const NetServiceSglRequest &req,
        const NetCallback *done);

    /**
     * @brief Post a single side read request to peer, no NetCallback at peer will be triggered. Poll mode:worker
     * poll(default mode, support sync/async with sem), self poll(support sync only)
     *
     * @param request      [in] request information, including 5 important variables, local/remote address/key and
     * size. You can free it after called NetServiceOneSideDoneHandler.
     * @param done         [in] NetCallback for async send to peer. If done is nullptr, the interface will block
     * until complete. And it means if set done is nullptr in worker callback context, the interface will block
     * worker thread.
     *
     * @return 0 if done is nullptr, it means the operation is done successfully. Otherwise, it means the operation
     * is in progress.
     */
    VIRTUAL SerResult Read(const NetServiceRequest &request, const NetCallback *done);

    /**
     * @brief  Post multi single side read request to peer, no NetCallback at peer will be triggered. Poll
     * mode:worker poll(default mode, support sync/async with sem), self poll(support sync only)
     *
     * @param request      [in] request information, fill with local different MRs, send to the same remote MR by
     * local MRs sequence, you can free it after called NetServiceOneSideDoneHandler. rKey/rAddress do not need to
     * assign. Note that upCtxData is private variable for HCOM service, user do not fill it up.
     * @param done         [in] NetCallback for async send to peer. If done is nullptr, the interface will block
     * until complete. And it means if set done is nullptr in worker callback context, the interface will block
     * worker thread.
     *
     * @return 0 if done is nullptr, it means the operation is done successfully. Otherwise, it means the operation
     * is in progress.
     * return ERROR, done will not be invoked.
     *
     */
    VIRTUAL SerResult ReadSgl(const NetServiceSglRequest &request, const NetCallback *done);

    /**
     * @brief Post a single side write request to peer, no NetCallback at peer will be triggered. Poll mode:worker
     * poll(default mode, support sync/async with sem), self poll(support sync only)
     *
     * @param request      [in] request information, including 5 important variables, local/remote address/key and
     * size. You can free it after called NetServiceOneSideDoneHandler.
     * @param done         [in] NetCallback for async send to peer. If done is nullptr, the interface will block
     * until complete. And it means if set done is nullptr in worker callback context, the interface will block
     * worker thread.
     *
     * @return 0 if done is nullptr, it means the operation is done successfully. Otherwise, it means the operation
     * is in progress.
     */
    VIRTUAL SerResult Write(const NetServiceRequest &request, const NetCallback *done);

    /**
     * @brief  Post multi single side write request to peer, no NetCallback at peer will be triggered. Poll
     * mode:worker poll(default mode, support sync/async with sem), self poll(support sync only)
     *
     * @param request      [in] request information, fill with local different MRs, send to the same remote MR by
     * local MRs sequence, you can free it after called NetServiceOneSideDoneHandler. rKey/rAddress do not need to
     * assign. Note that upCtxData is private variable for HCOM service, user do not fill it up.
     * @param done         [in] NetCallback for async send to peer. If done is nullptr, the interface will block
     * until complete. And it means if set done is nullptr in worker callback context, the interface will block
     * worker thread.
     *
     * @return 0 if done is nullptr, it means the operation is done successfully. Otherwise, it means the operation
     * is in progress.
     * return ERROR, done will not be invoked.
     *
     */
    VIRTUAL SerResult WriteSgl(const NetServiceSglRequest &request, const NetCallback *done);

    /**
     * @brief Only supported multi_rail view, post a single side read request to peer, no NetCallback at peer will be
     * triggered. Poll mode:worker poll(default mode, support sync/async with sem), self poll(support sync only)
     *
     * @param request      [in] request information, including 5 important variables, local/remote address/key and
     * size. You can free it after called NetServiceOneSideDoneHandler.
     * @param serviceMrKey [in] serviceMrKey means peer memory region keys
     * @param done         [in] NetCallback for async send to peer. If done is nullptr, the interface will block
     * until complete. And it means if set done is nullptr in worker callback context, the interface will block
     * worker thread.
     *
     * @return 0 if done is nullptr, it means the operation is done successfully. Otherwise, it means the operation
     * is in progress.
     */
    VIRTUAL SerResult Read(const NetServiceRequest &request, NetServiceMemoryRegionKey &serviceMrKey,
        const NetCallback *done);

    /**
     * @brief  Only supported multi_rail view, post multi single side read request to peer, no NetCallback at peer will
     * be triggered. Poll mode:worker poll(default mode, support sync/async with sem), self poll(support sync only)
     *
     * @param request      [in] request information, fill with local different MRs, send to the same remote MR by
     * local MRs sequence, you can free it after called NetServiceOneSideDoneHandler. rKey/rAddress do not need to
     * assign. Note that upCtxData is private variable for HCOM service, user do not fill it up.
     * @param serviceMrKey [in] serviceMrKey means peer memory region keys
     * @param done         [in] NetCallback for async send to peer. If done is nullptr, the interface will block
     * until complete. And it means if set done is nullptr in worker callback context, the interface will block
     * worker thread.
     *
     * @return 0 if done is nullptr, it means the operation is done successfully. Otherwise, it means the operation
     * is in progress.
     * return ERROR, done will not be invoked.
     *
     */
    VIRTUAL SerResult ReadSgl(const NetServiceSglRequest &request, NetServiceMemoryRegionKey &serviceMrKey,
        const NetCallback *done);

    /**
     * @brief Only supported multi_rail view, post a single side write request to peer, no NetCallback at peer will be
     * triggered. Poll mode:worker poll(default mode, support sync/async with sem), self poll(support sync only)
     *
     * @param request      [in] request information, including 5 important variables, local/remote address/key and
     * size. You can free it after called NetServiceOneSideDoneHandler.
     * @param serviceMrKey [in] serviceMrKey means peer memory region keys
     * @param done         [in] NetCallback for async send to peer. If done is nullptr, the interface will block
     * until complete. And it means if set done is nullptr in worker callback context, the interface will block
     * worker thread.
     *
     * @return 0 if done is nullptr, it means the operation is done successfully. Otherwise, it means the operation
     * is in progress.
     */
    VIRTUAL SerResult Write(const NetServiceRequest &request, NetServiceMemoryRegionKey &serviceMrKey,
        const NetCallback *done);

    /**
     * @brief Only supported multi_rail view, post multi single side write request to peer, no NetCallback at peer will
     * be triggered. Poll mode:worker poll(default mode, support sync/async with sem), self poll(support sync only)
     *
     * @param request      [in] request information, fill with local different MRs, send to the same remote MR by
     * local MRs sequence, you can free it after called NetServiceOneSideDoneHandler. rKey/rAddress do not need to
     * assign. Note that upCtxData is private variable for HCOM service, user do not fill it up.
     * @param serviceMrKey [in] serviceMrKey means peer memory region keys
     * @param done         [in] NetCallback for async send to peer. If done is nullptr, the interface will block
     * until complete. And it means if set done is nullptr in worker callback context, the interface will block
     * worker thread.
     *
     * @return 0 if done is nullptr, it means the operation is done successfully. Otherwise, it means the operation
     * is in progress.
     * return ERROR, done will not be invoked.
     *
     */
    VIRTUAL SerResult WriteSgl(const NetServiceSglRequest &request, NetServiceMemoryRegionKey &serviceMrKey,
        const NetCallback *done);

    /**
     * @brief Estimated Encrypt length for input raw len
     *
     * @param rawLen       [in] raw length before encrypt
     *
     * @return the length after encrypt, 0 means failed
     */
    VIRTUAL uint64_t EstimatedEncryptLen(uint64_t rawLen);

    /**
     * @brief Encrypt data
     *
     * @param rawData      [in] raw data before encrypt
     * @param rawLen       [in] raw data length before encrypt
     * @param cipher       [out] cipher data after encrypt
     * @param cipherLen    [out] cipher data length after encrypt
     *
     * @return 0 if success
     */
    VIRTUAL NResult Encrypt(const void *rawData, uint64_t rawLen, void *cipher, uint64_t &cipherLen);

    /**
     * @brief Estimate Decrypt length
     *
     * @param cipherLen    [in] cipher len before decrypt
     *
     * @return the raw length after decrypt, 0 means failed
     */
    VIRTUAL uint64_t EstimatedDecryptLen(uint64_t cipherLen);

    /**
     * @brief Decrypt data
     *
     * @param cipher       [in] cipher data after encrypt
     * @param cipherLen    [in] cipher data length after encrypt
     * @param rawData      [out] raw data before encrypt
     * @param rawLen       [out] raw data length before encrypt
     *
     * @return 0 if success
     */
    VIRTUAL NResult Decrypt(const void *cipher, uint64_t cipherLen, void *rawData, uint64_t &rawLen);

    /**
     * @brief Send shm fds, only shm protocol support
     *
     * @param fds          [in] fds to send
     * @param len          [in] fds count to send
     *
     * @return 0 if success
     */
    NResult SendFds(int fds[], uint32_t len);

    /**
     * @brief Receive shm fds, only shm protocol support
     *
     * @param fds          [out] fds to be received
     * @param len          [in] fds count to be received
     * @param timeoutSec   [in] timeout in second for receive. -1 is never timeout
     *
     * @return 0 if success
     */
    NResult ReceiveFds(int fds[], uint32_t len, int32_t timeoutSec);

    /**
     * @brief Get remote uds ids include pid uid gid, only support in oob server and when oob type is uds
     *
     * @param idInfo       [out] remote uds ids
     */
    NResult GetRemoteUdsIdInfo(NetUdsIdInfo &idInfo);

    /**
     * @brief close all connections in channel
     */
    VIRTUAL void Close();

    /**
     * @brief Set the timeout of the one side operation, include read/readSgl/write/writeSgl ...etc
     *
     * 1. timeout = 0: return immediately
     * 2. timeout < 0: never timeout, usually set to -1
     * 3. timeout > 0: second precision timeout.
     * Default is -1.
     *
     * @param timeout in seconds.
     */
    void SetOneSideTimeout(int16_t timeout);

    /**
     * @brief Set the timeout of the two side operation, include send/sendRaw/sendRawSgl/SyncCall/AsyncCall ...etc
     *
     * 1. timeout = 0: return immediately
     * 2. timeout < 0: never timeout, usually set to -1
     * 3. timeout > 0: second precision timeout.
     * Default is -1.
     *
     * @param timeout in seconds.
     */
    void SetTwoSideTimeout(int16_t timeout);

    /**
     * @brief config flow control
     */
    SerResult ConfigFlowControl(ChannelFlowCtrlOptions &options);

    /**
     * @brief Get channel type index
     */
    NetChTypeIndex TypeIndex() const;

    /**
     * @brief Get channel callback type
     */
    NetChannelCBType CBType() const;

    /**
     * @brief Get channel callback type
     */
    bool IsSelfPoll() const;

    /**
     * @brief Set the upper context, which could be used store user data pointer and read it when handler called
     */
    void UpCtx(uint64_t ctx);

    /**
     * @brief Get the upper context
     */
    uint64_t UpCtx() const;

    /**
     * @brief Get channel id
     */
    uint64_t Id() const;

    /**
     * @brief Get the local ip
     */
    uint32_t LocalIp() const;

    /**
     * @brief Get the payload
     */
    std::string PeerConnectPayload() const;

    /**
     * @brief Get the broken policy
     */
    NetChannelBrokenPolicy BrokenPolicy();

    /**
     * @brief Get the maximum using count in sending queues
     */
    inline uint32_t GetMaxSendQueueCount()
    {
        uint32_t res = 0;
        for (auto i = 0; i < mEpInfo->epSize; i++) {
            auto tmp = mEpInfo->epArr[i]->GetSendQueueCount();
            res = res >= tmp ? res : tmp;
        }
        return res;
    }

    /**
     * @brief Get ip and port of peer
     */
    inline bool GetPeerIpPort(std::string &ip, uint16_t &port) const
    {
        for (auto i = 0; i < mEpInfo->epSize; i++) {
            if (NN_LIKELY(mEpInfo->epArr[i]->GetPeerIpPort(ip, port))) {
                return true;
            }
        }

        return false;
    }

    inline void SetMemAllocMap(NetServiceMemoryRegionKey &serviceMrKey)
    {
        mServiceMrKey = serviceMrKey;
    }

    DEFINE_RDMA_REF_COUNT_FUNCTIONS

protected:
    enum ServiceEpState : uint16_t {
        SER_EP_ESTABLISHED = 0,
        SER_EP_BROKEN = 1,
        SER_EP_ESTABLISHED_OCCUPIED = 2,
        SER_EP_ESTABLISHED_UNOCCUPIED = 3,
    };

    enum ServiceChannelState : uint16_t {
        CH_NEW = 0,
        CH_ESTABLISHED = 1,
        CH_CLOSE = 2,
        CH_DESTROY = 3,
    };

    struct EpInfo {
        NetAtomicState<ServiceEpState> epState[NN_NO16] {}; /* state of eps */
        NetEndpoint *epArr[NN_NO16] {};                     /* endpoints for data transfer */
        uint16_t epSize = 0;

        EpInfo() = default;
    };

    struct ChannelInfo {
        uint16_t channelSize = 0;
        NetAtomicState<ServiceChannelState> channelState[NN_NO16]{}; /* state of channels */
        NetChannel *channelArr[NN_NO16]{};                     /* channels for data transfer */

        ChannelInfo() = default;
    };

protected:
    NetChannel(uint64_t id, bool selfPoll, bool reconnect, NetServiceConnectOptions &options)
        : mSelfPoll(selfPoll), mId(id), mReconnectChannel(reconnect), mConnOptions(options)
    {
        if (options.flags == NET_EP_SELF_POLLING || options.flags == NET_EP_EVENT_POLLING) {
            mRespOriginalSeqNo = true;
        }
        OBJ_GC_INCREASE(NetChannel);
        mChState.Set(CH_NEW);
    }

    virtual ~NetChannel()
    {
        UnInitialize();
        OBJ_GC_DECREASE(NetChannel);
    }

    SerResult InitializeEp(std::vector<NetEndpointPtr> &ep);

    SerResult Initialize(std::vector<NetEndpointPtr> &ep, uintptr_t ctxStore, uintptr_t periodicMgr);

    void ForceUnInitialize();

    void UnInitialize();

    std::string ToString();

    void SetEpUpCtx();

    void UnSetEpUpCtx();

    void SetEpBroken(uint32_t index) const;

    bool NeedProcessBroken();

    void RecoverProcessBroken();

    bool AllEpBroken();

    bool AllEpEstablished();

    SerResult ResponseWorkerPollEp(uintptr_t rspCtx, NetEndpoint *&ep);

    SerResult NextWorkerPollEp(NetEndpoint *&ep);

    SerResult AcquireSelfPollEp(NetEndpoint *&ep, uint32_t &index, int16_t timeout);

    void ReleaseSelfPollEp(uint32_t index);

    uint32_t SelfPollNextSeqNo();

    void ReconnectOobInfo(const std::string &ipOrName, uint16_t port);

    void Payload(const std::string &payload);

    void Uuid(const std::string &uuid);

    void BrokenInfo(uint16_t window, NetChannelBrokenPolicy policy, const NetServiceChannelBrokenHandler &broken);

    bool InvokeNewChannelCb();

    void InvokeChannelBrokenCb(NetChannelPtr &channel);

    uint16_t GetDelayEraseTime();

    void ProcessIoInBroken();

    SerResult FlowControl(uint64_t size, int16_t timeout, uint64_t timestamp);

    SerResult SendInner(const NetServiceOpInfo &opInfo, const NetServiceMessage &message, const NetCallback *done,
        uintptr_t rspCtx = 0);

    SerResult SendRawInner(const NetServiceMessage &req, const NetCallback *done, uintptr_t rspCtx);

    SerResult SendRawSglInner(const NetServiceSglRequest &req, const NetCallback *done, uintptr_t rspCtx);

    SerResult SyncCallInner(const NetServiceOpInfo &reqOpInfo, const NetServiceMessage &req,
        NetServiceOpInfo &rspOpInfo, NetServiceMessage &rsp);

    SerResult AsyncCallInner(const NetServiceOpInfo &opInfo, const NetServiceMessage &req, const NetCallback *done);

    SerResult SyncCallRawInner(const NetServiceMessage &req, NetServiceMessage &rsp);

    SerResult AsyncCallRawInner(const NetServiceMessage &req, const NetCallback *done);

    SerResult SyncCallRawSglInner(const NetServiceSglRequest &req, NetServiceMessage &rsp);

    SerResult AsyncCallRawSglInner(const NetServiceSglRequest &req, const NetCallback *done);

    SerResult SyncRndvCallInner(const NetServiceOpInfo &reqOpInfo, const NetServiceRequest &req,
        NetServiceOpInfo &rspOpInfo, NetServiceMessage &rsp);

    SerResult AsyncRndvCallInner(const NetServiceOpInfo &reqOpInfo, const NetServiceRequest &req,
        const NetCallback *done);

    SerResult SyncRndvSglCallInner(const NetServiceOpInfo &reqOpInfo, const NetServiceSglRequest &req,
        NetServiceOpInfo &rspOpInfo, NetServiceMessage &rsp);

    SerResult AsyncRndvSglCallInner(const NetServiceOpInfo &reqOpInfo, const NetServiceSglRequest &req,
        const NetCallback *done);

    SerResult ReadInner(const NetServiceRequest &request, const NetCallback *done);

    SerResult ReadSglInner(const NetServiceSglRequest &request, const NetCallback *done);

    SerResult WriteInner(const NetServiceRequest &request, const NetCallback *done);

    SerResult WriteSglInner(const NetServiceSglRequest &request, const NetCallback *done);

    SerResult PrepareTimerContext(NetCallback *cb, int16_t timeout, TimerContext &context);

    void DestroyTimerContext(TimerContext &context);

    SerResult SyncSendWithWorkerPoll(NetEndpoint *&ep, const NetServiceOpInfo &opInfo, const NetServiceMessage &message,
        uintptr_t rspCtx);

    SerResult SendWithSelfPoll(const NetServiceOpInfo &opInfo, const NetServiceMessage &message,
        const NetCallback *done, uintptr_t rspCtx);

    SerResult SyncSendRawWithWorkerPoll(NetEndpoint *&ep, const NetServiceMessage &message, uint32_t userSeqNo,
        uintptr_t rspCtx);

    SerResult SendRawWithSelfPoll(const NetServiceMessage &message, const NetCallback *done, uintptr_t rspCtx);

    SerResult SyncSendRawSglWithWorkerPoll(NetEndpoint *&ep, const NetServiceSglRequest &message, uint32_t userSeqNo,
        uintptr_t rspCtx);

    SerResult SendRawSglWithSelfPoll(const NetServiceSglRequest &message, const NetCallback *done, uintptr_t rspCtx);

    SerResult SyncCallWithSelfPoll(const NetServiceOpInfo &opInfo, const NetServiceMessage &req,
        NetServiceOpInfo &rspOpInfo, NetServiceMessage &rsp);

    SerResult SyncCallRawWithSelfPoll(const NetServiceMessage &req, NetServiceMessage &rsp);

    SerResult SyncCallRawSglWithSelfPoll(const NetServiceSglRequest &req, NetServiceMessage &rsp);

    SerResult SyncReadWithWorkerPoll(NetEndpoint *&ep, const NetServiceRequest &request);

    SerResult SyncReadWithSelfPoll(const NetServiceRequest &request);

    SerResult SyncReadSglWithWorkerPoll(NetEndpoint *&ep, const NetServiceSglRequest &request);

    SerResult SyncReadSglWithSelfPoll(const NetServiceSglRequest &request);

    SerResult SyncWriteWithWorkerPoll(NetEndpoint *&ep, const NetServiceRequest &request);

    SerResult SyncWriteWithSelfPoll(const NetServiceRequest &request);

    SerResult SyncWriteSglWithWorkerPoll(NetEndpoint *&ep, const NetServiceSglRequest &request);

    SerResult SyncWriteSglWithSelfPoll(const NetServiceSglRequest &request);

    inline void SetDevIndex(uint8_t devIndex)
    {
        mDevIndex = devIndex;
    }

    inline uint8_t GetDevIndex()
    {
        return mDevIndex;
    }

    inline void SetPeerDevIndex(uint8_t peerDevIndex)
    {
        mPeerDevIndex = peerDevIndex;
    }

    inline uint8_t GetPeerDevIndex()
    {
        return mPeerDevIndex;
    }

    inline void SetBandWidth(uint8_t bandWidth)
    {
        mBandWidth = bandWidth;
    }

    inline uint8_t GetBandWidth()
    {
        return mBandWidth;
    }

    /**
     * @brief Get state, don't change it which could leading to uncertain behavior
     */
    inline NetAtomicState<ServiceChannelState> &GetState()
    {
        return mChState;
    }

protected:
    /* cache line 64B start */
    EpInfo *mEpInfo = nullptr;       /* endpoint information */
    uintptr_t mCtxStore = NN_NO0;    /* store seq no ctx */
    uintptr_t mPeriodicMgr = NN_NO0; /* timeout periodic manager */
    int16_t mOneSideTimeout = -1;    /* default one side timeout */
    int16_t mTwoSideTimeout = -1;    /* default two side timeout */
    uint16_t mMsgRRIndex = 0;        /* index for choosing to which ep to transfer */
    bool mSelfPoll = false;          /* self thread poll or worker poll */
    bool mRespOriginalSeqNo = false; /* for mark seq no */
    uint64_t mUpCtx = 0;             /* context of this channel */
    uintptr_t mRateLimit = 0;        /* 0 means close, otherwise means open */
    uintptr_t mTimerList = 0;        /* record timer ctx by list */
    uint32_t mSeqNo = 1;             /* for self polling simplified usage */
    DEFINE_RDMA_REF_COUNT_VARIABLE
    // rsv 8 Byte
    /* cache line 64B end */

    uint64_t mId = 0;                                    /* channel uid */
    bool mClient = false;                                /* client support reconnect */
    bool mReconnectChannel = false;                      /* mark tmp channel reconnect type */
    bool mReconnectSuccess = false;                      /* reconnect success do not erase channel from map */
    uint16_t mConnectWindow = NN_NO60;                   /* keep channel alive for reconnect */
    uint16_t mOobPort = 0;                               /* record oob port for reconnect, exist in client */
    uint32_t mLocalIp = 0;                               /* record for generate uuid */
    uint64_t mReconnectTimeout = 0;                      /* if current > reconnect timeout, reject invoke reconnect */
    std::string mOobIpOrName;                            /* record oobIpOrName port for reconnect, exist in client */
    std::string mPayload;                                /* payload info */
    std::string mUuid;                                   /* uuid */
    NetServiceConnectOptions mConnOptions {};            /* connection option */
    ConnectTimestamp mConnectTimestamp {};               /* connection timestamp */
    NetChannelBrokenPolicy mChBrokenPolicy = BROKEN_ALL; /* broken policy set by client */
    NetServiceChannelBrokenHandler mChBrokenHandler = nullptr; /* channel broken handler */
    std::mutex mMgrMutex;                                      /* ch state manager mutex */
    NetAtomicState<ServiceChannelState> mChState;              /* ch state */
    std::atomic_bool mBrokenProcessed { false };               /* avoid race condition */
    uint8_t mDevIndex = 0;
    uint8_t mPeerDevIndex = 0;
    uint8_t mBandWidth = 0;
    NetServiceMemoryRegionKey mServiceMrKey;

    friend class NetService;
    friend class NetServiceDefaultImp;
    friend class NetServiceMultiRailImp;
    friend class ServiceNetDriverManager;
    friend class MultiRailNetChannel;
    friend class NetServiceContext;
    friend class NetServiceRndvContext;
    friend class NetServiceTimer;
    friend class NetPeriodicManager;
};

inline uint64_t NetChannel::EstimatedEncryptLen(uint64_t rawLen)
{
    NN_ASSERT_LOG_RETURN(mEpInfo != nullptr, 0)
    NN_ASSERT_LOG_RETURN(mEpInfo->epArr[0] != nullptr, 0)
    return mEpInfo->epArr[0]->EstimatedEncryptLen(rawLen);
}

inline NResult NetChannel::Encrypt(const void *rawData, uint64_t rawLen, void *cipher, uint64_t &cipherLen)
{
    NN_ASSERT_LOG_RETURN(mEpInfo != nullptr, SER_ERROR)
    NN_ASSERT_LOG_RETURN(mEpInfo->epArr[0] != nullptr, SER_ERROR)
    return mEpInfo->epArr[0]->Encrypt(rawData, rawLen, cipher, cipherLen);
}

inline uint64_t NetChannel::EstimatedDecryptLen(uint64_t cipherLen)
{
    NN_ASSERT_LOG_RETURN(mEpInfo != nullptr, 0)
    NN_ASSERT_LOG_RETURN(mEpInfo->epArr[0] != nullptr, 0)
    return mEpInfo->epArr[0]->EstimatedDecryptLen(cipherLen);
}

inline NResult NetChannel::Decrypt(const void *cipher, uint64_t cipherLen, void *rawData, uint64_t &rawLen)
{
    NN_ASSERT_LOG_RETURN(mEpInfo != nullptr, SER_ERROR)
    NN_ASSERT_LOG_RETURN(mEpInfo->epArr[0] != nullptr, SER_ERROR)
    return mEpInfo->epArr[0]->Decrypt(cipher, cipherLen, rawData, rawLen);
}

inline NetChTypeIndex NetChannel::TypeIndex() const
{
    return mConnOptions.index;
}

inline NetChannelCBType NetChannel::CBType() const
{
    return mConnOptions.cbType;
}

inline bool NetChannel::IsSelfPoll() const
{
    return mSelfPoll;
}

inline void NetChannel::UpCtx(uint64_t ctx)
{
    mUpCtx = ctx;
}

inline uint64_t NetChannel::UpCtx() const
{
    return mUpCtx;
}

inline uint64_t NetChannel::Id() const
{
    return mId;
}

inline void NetChannel::SetEpUpCtx()
{
    for (uint16_t i = 0; i < mEpInfo->epSize; i++) {
        Ep2ChUpCtx ctx(NN_NO1, reinterpret_cast<uint64_t>(this), i);
        mEpInfo->epArr[i]->UpCtx(ctx.wholeUpCtx);
    }
}

inline void NetChannel::UnSetEpUpCtx()
{
    for (uint16_t i = 0; i < mEpInfo->epSize; i++) {
        mEpInfo->epArr[i]->UpCtx(NN_NO0);
    }
}

inline bool NetChannel::AllEpEstablished()
{
    for (uint16_t i = 0; i < mEpInfo->epSize; i++) {
        if (mEpInfo->epState[i].Compare(SER_EP_BROKEN) || mEpInfo->epArr[i]->State().Compare(NEP_BROKEN)) {
            return false;
        }
    }

    return true;
}

inline SerResult NetChannel::ResponseWorkerPollEp(uintptr_t rspCtx, NetEndpoint *&ep)
{
    if (NN_UNLIKELY(!mChState.Compare(CH_ESTABLISHED))) {
        NN_LOG_ERROR("Channel state not established " << mChState.Get());
        return SER_NOT_ESTABLISHED;
    }

    uint32_t epIndex = rspCtx >> NN_NO32;
    if (NN_UNLIKELY(epIndex >= mEpInfo->epSize)) {
        NN_LOG_ERROR("Invalid ep index " << epIndex << " over ep size " << mEpInfo->epSize);
        return SER_INVALID_PARAM;
    }

    if (NN_UNLIKELY(mEpInfo->epState[epIndex].Compare(SER_EP_BROKEN))) {
        NN_LOG_ERROR("Ep broken of channel id " << Id() << " , select response ep fail");
        return SER_NOT_ESTABLISHED;
    }

    ep = mEpInfo->epArr[epIndex];
    if (NN_UNLIKELY(ep == nullptr)) {
        NN_LOG_ERROR("Channel Id " << mId << " ep invalid");
        return SER_NOT_ESTABLISHED;
    }
    return SER_OK;
}

inline SerResult NetChannel::NextWorkerPollEp(NetEndpoint *&ep)
{
    if (NN_UNLIKELY(!mChState.Compare(CH_ESTABLISHED))) {
        NN_LOG_ERROR("Channel state not established " << mChState.Get());
        return SER_NOT_ESTABLISHED;
    }

    uint16_t tmpIndex = __sync_fetch_and_add(&mMsgRRIndex, 1) % mEpInfo->epSize;
    uint16_t count = 0;

    while (mEpInfo->epState[tmpIndex].Compare(SER_EP_BROKEN) && count < mEpInfo->epSize) {
        tmpIndex = (tmpIndex + NN_NO1) % mEpInfo->epSize;
        count++;
    }

    if (NN_UNLIKELY(count > mEpInfo->epSize)) {
        NN_LOG_ERROR("Channel Id " << mId << " all ep broken");
        return SER_NOT_ESTABLISHED;
    }

    ep = mEpInfo->epArr[tmpIndex];
    if (NN_UNLIKELY(ep == nullptr)) {
        NN_LOG_ERROR("Channel Id " << mId << " ep invalid");
        return SER_NOT_ESTABLISHED;
    }
    return SER_OK;
}

inline void NetChannel::ReleaseSelfPollEp(uint32_t index)
{
    if (NN_UNLIKELY(index >= mEpInfo->epSize)) {
        NN_LOG_ERROR("Invalid index to release self poll ep in channel " << mId);
        return;
    }

    if (!mEpInfo->epState[index].CAS(SER_EP_ESTABLISHED_OCCUPIED, SER_EP_ESTABLISHED_UNOCCUPIED)) {
        NN_LOG_ERROR("Channel id " << mId << " failed to release self poll ep, state " <<
            mEpInfo->epState[index].Get());
    }
}

inline uint32_t NetChannel::SelfPollNextSeqNo()
{
    /* reserve 1 bit for mark send/rsp */
    uint32_t tmpSeqNo = __sync_fetch_and_add(&mSeqNo, 1);
    /* In order to make sure the netSeqNo.wholeSeq is not zero, and since only the lower 24 bits will be assigned to
       netSeqNo.realSeq, tmpSeqNo need to be ensured that lower 24 bits are not zero. */
    if (NN_UNLIKELY((tmpSeqNo & 0x00FFFFFF) == 0)) {
        tmpSeqNo = __sync_fetch_and_add(&mSeqNo, 1);
    }

    /* sell poll just set realSeq */
    NetSeqNo netSeqNo(0);
    netSeqNo.realSeq = tmpSeqNo;
    return netSeqNo.wholeSeq;
}

using NetServiceProtocol = NetDriverProtocol;
using NetServiceInfo = NetDriverDeviceInfo;
using NetServiceOobListenerOptions = NetOobListenerOptions;
using NetServiceOobUDSListenerOptions = NetOobUDSListenerOptions;
using NetServiceWorkingMode = NetDriverWorkingMode;
using NetServiceOobType = NetDriverOobType;
using NetServiceLBPolicy = NetDriverLBPolicy;

struct NetServiceOptions {
    char netDeviceIpMask[NN_NO256] {};             // ip masks for devices
    char netDeviceIpGroup[NN_NO1024] {};           // ip group for devices
    bool enableTls = true;                         // enable ssl
    TlsMode tlsMode = OPENSSL;                     // tls mode, default is openssl
    NetDriverSecType secType = NET_SEC_DISABLED;   // security type
    TlsVersion tlsVersion = TLS_1_3;               // tls version, default TLS1.3 (772)
    NetCipherSuite cipherSuite = AES_GCM_128;      // if tls enabled can set cipher suite, client and server should same
                                                   /* service setting */
    bool enableRndv = false;                       // enable rndv
    uint16_t connectWindow = NN_NO60;              // server side keep channel alive time in second
    uint32_t periodicThreadNum = NN_NO1;           // periodic check timeout req thread num, range[1, 4]
                                                   /* worker setting */
    bool dontStartWorkers = false;                 // start worker or not
    NetServiceWorkingMode mode = NET_BUSY_POLLING; // worker polling mode, could busy polling or event polling
    char workerGroups[NN_NO64] {};                 // worker groups, for example 1,3,3
    char workerGroupsCpuSet[NN_NO128] {};          // worker groups cpu set, for example 1-1,2-5,na
    char multiRailWorkerGroups[NN_NO256] {};       // worker groups, for example 1,3,3;1,2,3
    char multiRailWorkerGroupsCpuSet[NN_NO512] {}; // worker groups cpu set, for example 1-1,2-5,na;7-8,9-10,11-13
    // worker thread priority [-20,20], 20 is the lowest, -20 is the highest, 0 (default) means do not set priority
    int workerThreadPriority = 0;
		                                               /* connection setting */
    NetServiceOobType oobType = NET_OOB_TCP;       // oob type, tcp or UDS, UDS cannot accept remote connection
    NetServiceLBPolicy lbPolicy = NET_ROUND_ROBIN; // select worker load balance policy, default round-robin
    NetChTypeIndex maxTypeIndexSize = NN_NO1;      // service start() will check handlers with max size
    uint16_t magic = NN_NO256;                     // magic number for c/s connect validation
    uint8_t version = 0;                           // program version used by connect validation
                                                   /* heart beat setting */
    uint16_t heartBeatIdleTime = NN_NO60;          // heart beat idle time, in seconds
    uint16_t heartBeatProbeTimes = NN_NO7;         // heart beat probe times
    uint16_t heartBeatProbeInterval = NN_NO2;      // heart beat probe interval, in seconds
                                                   /* options for tcp protocol only */
    // timeout during io (s), it should be [-1, 1024], -1 means do not set, 0 means never timeout during io
    int16_t tcpUserTimeout = -1;
    bool tcpEnableNoDelay = true;                  // tcp TCP_NODELAY option, true in default
    bool tcpSendZCopy = false;                     // tcp whether copy request to inner memory, false in default
    /* The buffer sizes will be adjusted automatically when these two variables are 0, and the performance would be
     * better */
    uint16_t tcpSendBufSize = 0;                   // tcp connection send buffer size in kernel, in KB
    uint16_t tcpReceiveBufSize = 0;                // tcp connection send receive buf size in kernel, in KB
                                                   /* options for rdma protocol only */
    uint32_t mrSendReceiveSegCount = NN_NO8192;    // memory region segment count for two side operation
    uint32_t mrSendReceiveSegSize = NN_NO1024;     // data size of memory region segment
    uint16_t completionQueueDepth = NN_NO2048;     // completion queue size of rdma
    uint16_t maxPostSendCountPerQP = NN_NO64;      // max number request could issue
    uint16_t prePostReceiveSizePerQP = NN_NO64;    // pre post receive of qp
    uint16_t pollingBatchSize = NN_NO4;            // polling batch size for worker
    uint16_t eventPollingTimeout = NN_NO500;       // event polling timeout in us
    uint32_t qpSendQueueSize = NN_NO256;           // max send working request of qp for rdma
    uint32_t qpReceiveQueueSize = NN_NO256;        // max receive working request of qp for rdma
    bool enableMultiRail = false;                  // enable multi rail
    uint32_t minSegmentSize = NN_NO8192;           // multi rail split min size 8k~8M
    uint32_t maxConnectionNum = NN_NO250;          // max connection number
    uint64_t allocateResSize = NN_NO1048576;       // multiRail allocate resource size
    char oobPortRange[NN_NO16] {};                 // port range when enable port auto selection

    /**
     * @brief Set the ip mask for net devices, example: 192.168.0.1/24
     */
    bool SetNetDeviceIpMask(const std::string &mask);
    /**
     * @brief Set the ip mask for net devices, example: 192.168.0.1/24
     */
    bool SetNetDeviceIpGroup(const std::string &mask);

    /**
     * @brief Set worker groups, example: 1,3,4
     * meaning 3 groups for workers:
     * group0 has 1 workers
     * group1 has 3 workers
     * group2 has 4 workers
     */
    bool SetWorkerGroups(const std::string &value);

    /**
     * @brief Set worker groups cpu, example: 10-10,11-13,na
     * meaning 3 groups for workers:
     * group0 bind to cpu 10
     * group1 bind to cpu 11, 12, 13
     * group2 not bind to cpu
     */
    bool SetWorkerGroupsCpuSet(const std::string &value);
    /**
     * @brief MultiRail set worker groups, example: 1,3,4;1,2,1
     * meaning 2 groups for multiRail service:
     * group0 1,3,4
     * group1 1,2,1
     */
    bool SetMultiRailWorkerGroups(const std::string &value);
    /**
     * @brief MultiRail set worker groups cpu, example: 10-10,11-13,na;16-16,17-18,19-19
     * meaning 2 groups for multiRail service:
     * group0 10-10,11-13,na
     * group1 16-16,17-18,19-19
     */
    bool SetMultiRailWorkerGroupsCpuSet(const std::string &value);
} __attribute__((packed));

struct NetServiceOpHandlers {
    NetServiceReceivedHandler rcvHandler = nullptr;
    NetServiceSentHandler sentHandler = nullptr;
    NetServiceOneSideDoneHandler oneSideDoneHandler = nullptr;
};

using NetServiceRndvHandler = std::function<int(NetServiceRndvContext &)>;

class NetService {
public:
    /**
     * @brief Get a service instance by name
     *
     * @param t            [in] protocol of this service
     * @param name         [in] name of service to be created
     * @param startOobSvr  [in] start oob service or not
     *
     * @return Service instance is OK, otherwise return nullptr
     */
    static NetService *Instance(NetServiceProtocol t, const std::string &name, bool startOobSvr);

    /**
     * @brief Get a service instance by name in multiRail
     *
     * @param opt          [in] options of this service
     * @param t            [in] protocol of this service
     * @param name         [in] name of service to be created
     * @param startOobSvr  [in] start oob service or not
     *
     * @return Service instance is OK, otherwise return nullptr
     */
    static NetService *Instance(const NetServiceOptions &opt, NetServiceProtocol t, const std::string &name,
        bool startOobSvr);

    /**
     * @brief Destroy service instance by name
     *
     * @param name         [in] name of service to be created
     *
     * @return Destroy Service instance is OK, otherwise return error
     */
    static SerResult DestroyInstance(const std::string &name);

    /**
     * @brief Get trace information
     *
     * @return trace string
     */
    static std::string TraceLog();

    virtual ~NetService()
    {
        OBJ_GC_DECREASE(NetService);
    }

    /**
     * @brief Get the protocol of service
     */
    NetServiceProtocol Protocol() const;

    /**
     * @brief Set out of bound ip and port
     *
     * @param ip           [in]  ip address
     * @param port         [out] port
     */
    virtual void SetOobIpAndPort(const std::string &ip, uint16_t port);

    /**
     * @brief Get out of bound ip and port
     */
    bool GetOobIpAndPort(std::vector<std::pair<std::string, uint16_t>> &result);

    /**
     * @brief Add multiple oob listeners, if there is only one listener just use OobIpAndPort
     *
     * @param option       [in] listen options
     */
    void AddListener(const NetServiceOobListenerOptions &option);

    /**
     * @brief Set oob listener of uds type
     *
     * @param name     [in] name of uds listener
     *
     */
    void OobUdsName(const std::string &name);

    /**
     * @brief Add multiple oob uds listeners, if there is only one listener just use OobUdsName
     *
     * @param option       [in] option of uds listener option
     *
     */
    void AddOobUdsOptions(const NetServiceOobUDSListenerOptions &option);

    /**
     * @brief Start the net service
     *
     * @return 0 if successful
     */
    virtual SerResult Start(const NetServiceOptions &opt) = 0;

    /**
     * @brief Stop the net driver
     */
    virtual void Stop() = 0;

    /**
     * @brief Connect to server
     *
     * @param oobIpOrName  [in]  oob ip or name to connect, set ip for tcp and name for uds
     * @param oobPort      [in]  only need to set when tcp oob
     * @param payload      [in]  payload transferred to peer, will be got EP Connected NetCallback at server
     * @param ch           [out] connected channel
     * @param options      [in]  variable argument for specific use
     *
     * @return 0 successful
     */
    virtual SerResult Connect(const std::string &oobIpOrName, uint16_t oobPort, const std::string &payload,
        NetChannelPtr &ch, NetServiceConnectOptions options) = 0;

    /**
     * @brief Reconnect to server
     *
     * @param ch           [in] connected channel
     *
     * @return 0 successful
     */
    virtual SerResult Reconnect(NetChannelPtr &ch) = 0;

    /**
     * @brief Register a memory region, the memory will be allocated internally
     *
     * @param size         [in]  size of the memory region
     * @param mr           [out] memory region registered
     *
     * @return 0 successful
     */
    SerResult RegisterMemoryRegion(uint64_t size, NetMemoryRegionPtr &mr);

    /**
     * @brief Register a memory region, the memory need to be passed in
     *
     * @param address      [in]  the memory point need to be registered
     * @param size         [in]  size of the memory region
     * @param mr           [out] memory region registered
     *
     * @return 0 successful
     */
    SerResult RegisterMemoryRegion(uintptr_t address, uint64_t size, NetMemoryRegionPtr &mr);

    /**
     * @brief Multi Rail Register a memory region, the memory need to be passed in
     *
     * @param address      [in]  the memory point need to be registered
     * @param size         [in]  size of the memory region
     * @param serviceMrKey [out] Register memory region key array
     *
     * @return 0 successful
     */
    virtual SerResult RegisterMemoryRegion(uintptr_t address, uint64_t size, NetServiceMemoryRegionKey &serviceMrKey);

    /**
     * @brief Unregister the memory region
     * @param mr           [in] memory region registered
     *
     * @return 0 successful
     */
    void DestroyMemoryRegion(NetMemoryRegionPtr &mr);

    void DestroyChannel(NetChannelPtr &ch);

    /**
     * @brief Register NetCallback for new channel connected from client, only need to register at server side
     *
     * @param handler      [in] handler function
     */
    void RegisterNewChannelHandler(const NetServiceNewChannelHandler &handler);

    /**
     * @brief Register NetCallback for channel broken
     *
     * @param handler      [in] handler function
     * @param policy       [in] broken policy
     */
    void RegisterChannelBrokenHandler(const NetServiceChannelBrokenHandler &handler,
        const NetChannelBrokenPolicy policy);

    void RegisterIdleHandler(const NetServiceIdleHandler &h);

    /**
     * @brief Register NetCallback for channel type operation handlers
     *
     * @param i                    [in] channel type index
     * @param rcvHandler           [in] receive new msg handler
     */
    void RegisterOpReceiveHandler(NetChTypeIndex i, const NetServiceReceivedHandler &rcvHandler);

    /**
     * @brief Register NetCallback for channel type operation handlers
     *
     * @param i                    [in] channel type index
     * @param sentHandler          [in] send done handler
     */
    void RegisterOpSentHandler(NetChTypeIndex i, const NetServiceSentHandler &sentHandler);

    /**
     * @brief Register NetCallback for channel type operation handlers
     *
     * @param i                    [in] channel type index
     * @param oneSideDoneHandler   [in] one side done handler
     */
    void RegisterOpOneSideHandler(NetChTypeIndex i, const NetServiceOneSideDoneHandler &oneSideDoneHandler);

    /**
     * @brief Register allocate handler for RNDV allocate
     * @param handler             [in] allocate handler
     */
    void RegisterRndvAllocateHandler(const NetServiceRndvAllocateHander &handler);

    /**
     * @brief Register free handler for RNDV allocate
     * @param handler             [in] free handler
     */
    void RegisterRndvFreeHandler(const NetServiceRndvFreeHander &handler);

    /**
     * @brief Register free handler for RNDV allocate
     * @param handler             [in] free handler
     */
    void RegisterRndvHandler(const NetServiceRndvHandler &handler);

    /**
     * @brief Register callback for idle
     *
     * @param handler      [in] handler function
     */
    VIRTUAL void RegisterTLSCaCallback(const TLSCaCallback &cb);

    /**
     * @brief Register callback for idle
     *
     * @param handler      [in] handler function
     */
    VIRTUAL void RegisterTLSCertificationCallback(const TLSCertificationCallback &cb);

    /**
     * @brief Register callback for idle
     *
     * @param handler      [in] handler function
     */
    VIRTUAL void RegisterTLSPrivateKeyCallback(const TLSPrivateKeyCallback &cb);

    /**
     * @brief Register callback for create secure info
     *
     * @param provider      [in] provider function
     */
    void RegisterSecInfoProvider(const NetDriverEndpointSecInfoProvider &provider);

    /**
     * @brief Register callback for validate secure info from peer
     *
     * @param validator      [in] validator function
     */
    void RegisterSecInfoValidator(const NetDriverEndpointSecInfoValidator &validator);

    /**
     * @brief Get driver for more efficient communication
     *
     * @return driver smart ptr
     */
    const NetDriverPtr &GetDriver() const;

    /**
     * @brief Get the name of service
     */
    const std::string &Name() const;

    /**
     * @brief Get the result indicates whether service is stopped
     */
    bool IsStarted() const;

    DEFINE_RDMA_REF_COUNT_FUNCTIONS
protected:
    SerResult ValidateHandler(const NetServiceOptions &opt, NetChTypeIndex maxSize);

    NetService(const std::string &name, NetDriver *driver, bool startOob)
        : mDriverPtr(driver), mName(name), mStartOob(startOob)
    {
        OBJ_GC_INCREASE(NetService);
    }

protected:
    NetServiceIdleHandler mIdleHandler = nullptr;
    NetServiceOpHandlers mOpHandlers[NN_NO16];
    NetChTypeIndex mMaxTypeIndexSize = NN_NO1;

    NetServiceNewChannelHandler mChNewHandler = nullptr;
    NetServiceChannelBrokenHandler mChBrokenHandler = nullptr;

    NetServiceRndvAllocateHander mAllocateHandler = nullptr;
    NetServiceRndvFreeHander mFreeHandler = nullptr;
    NetServiceRndvHandler mRndvHandler = nullptr;

    NetDriverEndpointSecInfoProvider mSecInfoProvider = nullptr;
    NetDriverEndpointSecInfoValidator mSecInfoValidator = nullptr;

    NetDriverPtr mDriverPtr = nullptr;

    std::string mName;
    bool mStartOob = false;
    bool mStarted = false;
    uint8_t mIndex = 0;
    NetChannelBrokenPolicy mChBrokenPolicy = BROKEN_ALL;
    std::string mOobIp;

private:
    static NetService *CreateMultiRailNetService(const NetServiceOptions &opt, NetServiceProtocol t,
                                               const std::string &name, bool startOobSvr);

private:
    static bool mEnableMultiRail;
    static uint8_t G_SERVICE_INDEX;
    static std::mutex G_MUTEX;
    static std::map<std::string, NetService *> G_SERVICE_MAP;
    DEFINE_RDMA_REF_COUNT_VARIABLE
};

inline NetServiceProtocol NetService::Protocol() const
{
    if (mDriverPtr == nullptr) {
        return UNKNOWN;
    }

    return mDriverPtr->Protocol();
}

inline const std::string &NetService::Name() const
{
    return mName;
}

inline bool NetService::IsStarted() const
{
    return mStarted;
}
}
}

#endif // OCK_HCOM_SERVICE_H