/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2021-2021. All rights reserved.
 * Author: bao, chen, fu, liu, ye
 */
#ifndef HCOM_HCOM_SERVICE_DEF_H
#define HCOM_HCOM_SERVICE_DEF_H

namespace ock {
namespace hcom {
class NetCallback;

/**
 * The meta info for one linked list,
 */
struct NetServiceSeqNoMap {
    std::unordered_map<uint32_t, NetCallback *> seqMap = {}; /* seq no map for save context */
    NetSpinLock lock {};                                     /* spin lock for seq no map */

    NetServiceSeqNoMap() = default;
};

enum NetServiceMapResult {
    SM_OK = 0,
    SM_EXIST = 1,
    SM_NOT_EXIST = 2,

    SM_BUFF
};

/**
 * A thread safe map with buckets for multiple threads cases,
 * used for call method saving context.
 */
template <uint16_t BUCKET_COUNT> class NetServiceContextMap {
public:
    NetServiceContextMap()
    {
        for (int i = 0; i < BUCKET_COUNT; i++) {
            mBuckets[i].seqMap.reserve(NN_NO8192);
        }
    }
    ~NetServiceContextMap() = default;

    /**
     * @brief Insert one item to map
     *
     * @param seqNo         [in] user define seq no
     * @param cb            [in] net callback address
     */
    inline NetServiceMapResult Emplace(uint32_t seqNo, NetCallback *cb)
    {
        NetServiceSeqNoMap *buckets = &mBuckets[seqNo % BUCKET_COUNT];

        buckets->lock.Lock();
        auto findItem = buckets->seqMap.find(seqNo);
        if (findItem != buckets->seqMap.end()) {
            buckets->lock.Unlock();
            return SM_EXIST;
        }

        buckets->seqMap.emplace(seqNo, cb);
        buckets->lock.Unlock();
        return SM_OK;
    }

    /**
     * @brief Get item and erase
     *
     * @param seqNo         [in] user define seq no
     * @param cb            [in] net callback address
     */
    inline NetServiceMapResult Fetch(uint32_t seqNo, NetCallback *&cb)
    {
        NetServiceSeqNoMap *buckets = &mBuckets[seqNo % BUCKET_COUNT];

        buckets->lock.Lock();
        auto findItem = buckets->seqMap.find(seqNo);
        if (findItem == buckets->seqMap.end()) {
            buckets->lock.Unlock();
            return SM_NOT_EXIST;
        }

        cb = findItem->second;
        buckets->seqMap.erase(seqNo);
        buckets->lock.Unlock();
        return SM_OK;
    }
    /**
     * @brief Delete one item from map
     *
     * @param seqNo         [in] user define seq no
     */
    inline void Erase(uint32_t seqNo)
    {
        NetServiceSeqNoMap *buckets = &mBuckets[seqNo % BUCKET_COUNT];

        buckets->lock.Lock();
        auto findItem = buckets->seqMap.find(seqNo);
        if (findItem == buckets->seqMap.end()) {
            buckets->lock.Unlock();
            return;
        }

        buckets->seqMap.erase(seqNo);
        buckets->lock.Unlock();
    }

    NetServiceContextMap(const NetServiceContextMap &) = delete;
    NetServiceContextMap(NetServiceContextMap &&) = delete;
    NetServiceContextMap &operator = (const NetServiceContextMap &) = delete;
    NetServiceContextMap &operator = (NetServiceContextMap &&) = delete;

private:
    NetServiceSeqNoMap mBuckets[BUCKET_COUNT]; /* buckets linked list */
};
}
}

#endif // HCOM_HCOM_SERVICE_DEF_H
